# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 15:43:05 2019

@author: Angshuman Deb
"""

from .multivariate_nataf import multivariate_nataf
import os
import platform
import numpy as np
import importlib.util
from scipy.stats import norm
import subprocess
import pickle
from openpyxl import load_workbook
import xlrd
import string
import eqsig
from datetime import datetime
from benedict import benedict
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib import cm
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class Utility:
    ####################################################################################################################
    # Namespace class with static methods
    ####################################################################################################################

    @staticmethod
    def str_replace(str_from, str_to, file_to_modify, path_of_file):
        if str_from is None:
            return
        file_to_modify = Utility.get_path(path_of_file, file_to_modify)
        with open(file_to_modify, 'r') as fid:
            lines = fid.readlines()
        line_inds = [ind for ind in range(len(lines)) if str_from in lines[ind]]
        for ind in line_inds:
            lines[ind] = lines[ind].replace(str_from, str_to)
        with open(file_to_modify, 'w') as fid:
            fid.writelines(lines)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def rgb(c):
        return colors.to_rgb(c)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def block_diag_3d(*args):
        depth = args[0].shape[2]
        shapes = np.array([a.shape[:-1] for a in args])
        out = np.zeros(tuple(np.sum(shapes, axis=0)) + (depth,))
        r, c = 0, 0
        for i, (rr, cc) in enumerate(shapes):
            out[r:r + rr, c:c + cc, :] = args[i]
            r += rr
            c += cc
        return out

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def is_list_of_same_elems(lst):
        return all(ele == lst[0] for ele in lst)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def plot_raw_ground_acc_time_history(gm_database_dir_path, **kwargs):
        rec_num = kwargs.get('rec_num', None)
        gm_file_names = kwargs.get('gm_file_names', None)
        scale_fac = kwargs.get('scale_fac', 1.0)
        if rec_num is not None and gm_file_names is None:
            # Extract rec_data
            rec_data_file_path = Utility.get_path(os.path.dirname(__file__), 'Rec_Data', 'rec_data.pickle')
            rec_data = Utility.pickle_load_dict(rec_data_file_path)
            file_name_1 = rec_data['file_name_1']
            file_name_2 = rec_data['file_name_2']
            file_name_vert = rec_data['file_name_vert']
            file_ind = rec_num - 1
            gm_file_names = [file_name_1[file_ind], file_name_2[file_ind], file_name_vert[file_ind]]
            gm_file_paths = [Utility.get_path(gm_database_dir_path, gm_file_name) for gm_file_name in gm_file_names]
        elif gm_file_names is not None and rec_num is None:
            gm_file_paths = [Utility.get_path(gm_database_dir_path, gm_file_name) for gm_file_name in gm_file_names]
        else:
            raise ValueError("either rec_num or gm_file_names should be not NoneType")
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        lc = kwargs.get('lc', 'black')
        max_marker = kwargs.get('max_marker', 'o')
        mec = kwargs.get('mec', 'red')
        ms = kwargs.get('ms', 5)
        lw = kwargs.get('lw', 0.8)
        fs = kwargs.get('fs', 8)
        txc = kwargs.get('txc', 'red')
        max_text_precision = kwargs.get('max_text_precision', 5.3)
        fig, axs = plt.subplots(len(gm_file_paths), 1)
        for itr in range(len(gm_file_paths)):
            gm, npts, dt = Utility.read_nga_peer_eq_at2_file(gm_file_paths[itr])
            if gm is not None:
                t = np.linspace(0.0, dt * npts, int(npts))
                axs[itr].plot(t, gm * scale_fac, '-', color=lc, linewidth=lw)
                axs[itr].minorticks_on()
                axs[itr].grid(True, which="major", alpha=grid_alpha)
                axs[itr].grid(True, which="minor", alpha=minor_grid_alpha)
                axs[itr].set_title(gm_file_names[itr] + f', scale: {scale_fac:.3}', fontsize=fs)
                axs[itr].tick_params(axis='x', labelsize=fs)
                axs[itr].tick_params(axis='y', labelsize=fs)
                Utility.mark_max_in_time_hist(axs[itr], gm * scale_fac, t, max_marker,
                                              markeredgecolor=mec,
                                              markersize=ms,
                                              precision=max_text_precision,
                                              units='g',
                                              textcolor=txc,
                                              fontsize=fs,
                                              ha='left',
                                              va='center')
            else:
                axs[itr].set_title('File NOT found', fontsize=fs)
        return fig, axs

    @staticmethod
    def fetch_selected_ground_motion_records(gm_database_dir_path, gm_file_names, scale_fac, target_dir_path,
                                             ai_end):
        time_hist_list = list()
        npts_list = list()
        dt_list = list()
        t_start_list = list()
        t_end_list = list()
        for (i, gm_file) in zip(list(range(1, len(gm_file_names) + 1)), gm_file_names):
            time_hist_temp, npts_temp, dt_temp = Utility.read_nga_peer_eq_at2_file(
                Utility.get_path(gm_database_dir_path, gm_file))
            time_hist_list.append(time_hist_temp)
            npts_list.append(npts_temp)
            dt_list.append(dt_temp)
            if npts_temp is not None:
                t_sig = eqsig.im.calc_sig_dur(eqsig.AccSignal(time_hist_temp * 9.81, dt_temp),
                                              start=0.001, end=ai_end,
                                              se=True)
            else:
                t_sig = [None, None]
            t_start_list.append(t_sig[0])
            t_end_list.append(t_sig[1])

        if npts_list[-1] is None:
            npts_list[-1] = npts_list[0]
            dt_list[-1] = dt_list[0]
            time_hist_list[-1] = time_hist_list[0] * 0.
            t_start_list[-1] = t_start_list[0]
            t_end_list[-1] = t_end_list[0]

        t_start = min(t_start_list)
        t_end = max(t_end_list)
        for i in range(len(gm_file_names)):
            npts_start = round(t_start / dt_list[i])
            npts_end = round(t_end / dt_list[i])
            time_hist = time_hist_list[i][int(npts_start):int(npts_end)]
            save_arr = np.vstack([npts_end - npts_start, dt_list[i], time_hist.reshape(len(time_hist), 1) * scale_fac])
            Utility.save_array_as_text_file(Utility.get_path(target_dir_path, f'gm_{i + 1}.txt'), save_arr)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def save_array_as_text_file(filepath, array, **kwargs):
        np.savetxt(Utility.get_path(filepath), array, **kwargs)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def read_nga_peer_eq_at2_file(file_path):
        file_path = file_path.replace('.at2', '.AT2')

        if not os.path.isfile(file_path):
            return None, None, None

        with open(file_path, 'r') as fid:
            skipheader = 0
            go_on = True
            while go_on:
                curr_line = fid.readline()
                words = curr_line.split()
                if all(Utility.is_numeric_string(word) for word in words):
                    go_on = False
                    num_cols = len(words)
                else:
                    skipheader += 1

        with open(file_path, 'r') as fid:
            lines = fid.readlines()

        line_index = -1
        go_on = True
        while go_on:
            curr_line = lines[line_index].strip().split()
            if len(curr_line) == num_cols:
                go_on = False
            else:
                line_index -= 1
        skipfooter = (-line_index) - 1

        time_hist = np.loadtxt(file_path, skiprows=skipheader, max_rows=len(lines) - skipheader - skipfooter).flatten()
        if line_index != -1:
            last_row = np.array(lines[line_index + 1].split()).astype(float)
            time_hist = np.hstack((time_hist, last_row))

        temp = [float(word) for word in lines[skipheader - 1].replace(',', ' ').replace(';', ' ').split() if
                Utility.is_numeric_string(word)]
        npts = int(max(temp))
        dt = float(min(temp))

        return time_hist, npts, dt

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_tag_of_token(name, token):
        name_parts = name.split('/')
        name_parts_with_token = [name_part for name_part in name_parts if token in name_part.split('_')]
        tags = list()
        for name_part_with_token in name_parts_with_token:
            parts_of_name_part_with_token = name_part_with_token.split('_')
            if '.' in parts_of_name_part_with_token[-1]:
                parts_of_name_part_with_token[-1] = parts_of_name_part_with_token[-1].split('.')[0]
            inds = [i for i in range(len(parts_of_name_part_with_token)) if parts_of_name_part_with_token[i] == token]
            tags.extend([parts_of_name_part_with_token[ind + 1] for ind in inds if
                         len(parts_of_name_part_with_token) != ind + 1
                         and Utility.is_numeric_string(parts_of_name_part_with_token[ind + 1])])
        if len(tags) > 0:
            return tags[0]
        else:
            return ''

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def exist_line_in_file(fid, line_to_test):
        line_to_test = line_to_test.strip('\n')
        pos = fid.tell()
        fid.seek(0)

        lines = fid.readlines()

        try:
            lines.index(f'{line_to_test}\n')
        except ValueError:
            try:
                lines.index(f'{line_to_test};\n')
            except ValueError:
                line_found = False
                fid.seek(pos)
                return line_found

        line_found = True
        fid.seek(pos)
        return line_found

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def pickle_dump_dict(file_path, dictionary):
        if type(dictionary) == benedict:
            dictionary = dict(dictionary)
        with open(file_path, 'wb') as f:
            pickle.dump(dictionary, f)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def pickle_load_dict(file_path):
        try:
            with open(file_path, 'rb') as f:
                dictionary = pickle.load(f)
            return dictionary
        except FileNotFoundError:
            raise FileNotFoundError

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def in_nd(a, b):

        a = list(a)
        b = list(b)

        a = [list(item) for item in a]
        b = [list(item) for item in b]

        to_return = [True if item in b else False for item in a]

        return np.array(to_return)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def combine_contents_of_files(dir_path, which_files, out_file_name, **kwargs):

        # which_files = ['all', 'startswith <tag>', 'contains <tag>', 'file_1', 'file_2', ...]

        file_ext = kwargs.get('file_ext', '.txt')
        delete_original = kwargs.get('delete_original', False)
        out_file_open_mode = kwargs.get('out_file_open_mode', 'w')

        files_to_cat = list()

        stuff_in_dir_path = os.listdir(dir_path)

        for file_names in which_files:
            if type(file_names) == str and file_names == 'all':
                files_to_cat.extend([stuff for stuff in stuff_in_dir_path if stuff.endswith(file_ext)])

            elif type(file_names) == str and file_names.startswith('startswith '):
                tag = file_names.split()[1]
                files_to_cat.extend(
                    [stuff for stuff in stuff_in_dir_path if stuff.startswith(tag) and stuff.endswith(file_ext)])

            elif type(file_names) == str and file_names.startswith('contains '):
                tags = file_names.split()[1:]
                files_to_cat.extend(
                    [stuff for stuff in stuff_in_dir_path if all(tag in stuff for tag in tags)
                     and stuff.endswith(file_ext)])

            elif type(file_names) == str:
                files_to_cat.append(file_names)

        files_to_cat = list(np.unique(files_to_cat))
        lines = Utility.get_lines_in_files(dir_path, files_to_cat)

        if len(files_to_cat) > 0:
            if '\\' in out_file_name or '/' in out_file_name:
                out_file_path = out_file_name
            else:
                out_file_path = Utility.get_path(dir_path, out_file_name)
            with open(out_file_path, out_file_open_mode) as fid:
                fid.writelines(lines)

        if len(files_to_cat) == 1 and files_to_cat[0] == out_file_name:
            delete_original = False

        if delete_original:
            [os.remove(Utility.get_path(dir_path, file)) for file in files_to_cat]

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_lines_in_files(dir_path, files_to_cat):
        lines = list()
        for file in files_to_cat:
            with open(Utility.get_path(dir_path, file), 'r') as fid:
                lines.extend(fid.readlines())
                lines[-1] = lines[-1].strip('\n')
                lines[-1] += '\n'

        return lines

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def make_array_2d(array):
        if array.ndim == 1:
            return array.reshape(1, array.shape[0])
        else:
            return array

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def make_array_3d(array):
        if array.ndim == 1:
            return array.reshape(1, array.shape[0], 1)
        elif array.ndim == 2:
            return array.reshape(array.shape[0], array.shape[1], 1)
        else:
            return array

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def write_tcl_input_file(fid, param_values, param_names):
        for itr in range(len(param_values)):
            if type(param_values[itr]) != str:
                fid.write(f'set {param_names[itr]} {param_values[itr]};\n')
            else:
                fid.write(f'set {param_names[itr]} "{param_values[itr]}";\n')

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_multi_level_iterator_cases(*args):
        # Input args should be lists or 1d array_like of unique numeric elements!
        vec = list()
        for arg in args:
            if type(arg) != list and type(arg) != np.ndarray:
                arg = [arg]
            arg = np.array(arg).reshape(len(arg))
            vec.append(arg)

        itr_range_size = [len(i) for i in vec]
        num_cases = np.prod(itr_range_size)
        itr_range_size_padded = [1, *itr_range_size, 1]
        cases = np.zeros((num_cases, len(itr_range_size)))
        curr_level = -1

        while curr_level >= -len(vec):
            temp = np.repeat(np.array(vec[curr_level]), np.prod(itr_range_size_padded[curr_level:]))
            temp = temp.reshape(len(temp), 1)
            temp = np.tile(temp, (np.prod(itr_range_size_padded[:curr_level - 1]), 1))
            cases[:, curr_level] = temp[:, 0]
            curr_level -= 1

        return cases

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_num(list_of_ints):
        return int(''.join([str(i) for i in list_of_ints]))

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_standard_beta_dist_params(mean, std, min_val, max_val):
        mu = mean / (max_val - min_val) - min_val / (max_val - min_val)
        sig = 1 / (max_val - min_val) * std
        a = -mu * np.divide(np.square(sig) + np.square(mu) - mu, np.square(sig))
        b = (np.square(sig) + np.square(mu) - mu) * np.divide(mu - 1, np.square(sig))
        return a, b

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_standard_normal_clips(myclip_a, myclip_b, my_mean, my_std):
        return (myclip_a - my_mean) / my_std, (myclip_b - my_mean) / my_std

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_my_normal_clips(a, b, my_mean, my_std):
        return a * my_std + my_mean, b * my_std + my_mean

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_lognormal_dist_params(mean, std, **kwargs):
        loc = kwargs.get('loc', np.abs(0.0 * mean))
        mean -= loc
        sig = np.sqrt(np.log(1 + np.square(std / np.abs(mean))))
        mu = np.log(mean) - 0.5 * np.square(sig)
        return sig, loc, np.exp(mu)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_all_dist_params(prob_dist):
        if prob_dist.dist.name == 'mixture':
            return prob_dist.args
        temp = prob_dist.dist.__dict__['_parse_args'](*prob_dist.args, **prob_dist.kwds)
        return [*temp[0], *temp[1:len(temp)]]

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def setup_rng(rng_seed):
        if (type(rng_seed) == str) and rng_seed.startswith('unique'):
            set_rng_seed = True
            temp = rng_seed.split(' ')
            if len(temp) > 1:
                seed_multiplier = int(float(temp[1]))
            else:
                seed_multiplier = 1
        else:
            set_rng_seed = False
            seed_multiplier = 1

        return set_rng_seed, seed_multiplier

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def import_attr_from_module(module_path, module_name, attr):
        module_path = Utility.get_path(module_path, module_name + '.py')
        spec = importlib.util.spec_from_file_location(module_name, module_path)
        temp = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(temp)
        return getattr(temp, attr)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_path(*args):
        args = [arg.replace('\\', '/') if isinstance(arg, str) else None for arg in args]
        args = list(filter(None, args))
        return '/'.join(args)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def is_numeric_string(n):
        try:
            # Type-casting the string to `float`.
            # If string is not a valid `float`, 
            # it'll raise `ValueError` exception
            float(n)
        except ValueError:
            return False
        return True

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def read_numpy_array_from_txt_file(file_path, **kwargs):

        # skiprows = any int, 0 (default), or 'find'

        skiprows = kwargs.get('skiprows', 0)
        comments = kwargs.get('comments', '#')

        if type(skiprows) == str and skiprows == 'find':
            with open(file_path, 'r') as fid:
                skiprows = 0
                go_on = True
                while go_on:
                    curr_line = fid.readline()
                    words = curr_line.split()
                    if all(Utility.is_numeric_string(word) for word in words) and len(words) > 0:
                        go_on = False
                    else:
                        skiprows += 1

        return Utility.make_array_2d(np.loadtxt(file_path, skiprows=skiprows, comments=comments))

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_indexed_matrix(m, ind):
        n = len(ind)
        inds = np.arange(m.shape[0])
        m_ind = np.in1d(inds, ind)
        m_ind = np.outer(m_ind, m_ind)
        return m[m_ind].reshape((n, n))

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def generate_random_dist_param_vals(prob_dist_list, corr_matrix, estimation_sample_size, dist_params_sample_size,
                                        **kwargs):

        rng_seed = kwargs.get('rng_seed', None)

        nd = len(prob_dist_list)
        out_list = list()
        m = multivariate_nataf(prob_dist_list, corr_matrix)
        m_rvs = m.rvs(size=(dist_params_sample_size, estimation_sample_size),
                      random_state=rng_seed,
                      method='mcs').reshape((dist_params_sample_size, estimation_sample_size, nd))
        m_rvs_mean = np.mean(m_rvs, axis=1)[:, np.newaxis, :]
        m_rvs_std = np.std(m_rvs, axis=1, ddof=1)
        m_rvs_temp = m_rvs - m_rvs_mean
        corr_matrix_realizations = (1 / (estimation_sample_size - 1) * (
            np.einsum('ijk,ijl->ikl', m_rvs_temp, m_rvs_temp))) / np.einsum('ij,ik->ijk', m_rvs_std, m_rvs_std)
        corr_matrix = corr_matrix[np.newaxis, :, :].repeat(dist_params_sample_size, axis=0)
        corr_matrix_realizations[corr_matrix == 0] = 0
        corr_matrix_realizations[corr_matrix == 1] = 1
        m_rvs_mean = m_rvs_mean.squeeze(axis=1)

        for itr in range(nd):
            temp = Utility.get_all_dist_params(prob_dist_list[itr])

            if prob_dist_list[itr].dist.name == 'beta':
                beta_a, beta_b = Utility.get_standard_beta_dist_params(m_rvs_mean[:, itr], m_rvs_std[:, itr], temp[2],
                                                                       temp[3] + temp[2])
                out_list.append(np.column_stack(
                    (beta_a, beta_b, [temp[2]] * dist_params_sample_size, [temp[3]] * dist_params_sample_size)))

            if prob_dist_list[itr].dist.name == 'lognorm':
                lognorm_zeta, lognorm_loc, lognorm_med = Utility.get_lognormal_dist_params(
                    m_rvs_mean[:, itr], m_rvs_std[:, itr], loc=[temp[1]] * dist_params_sample_size)
                out_list.append(np.column_stack((lognorm_zeta, lognorm_loc, lognorm_med)))

            if prob_dist_list[itr].dist.name == 'norm':
                out_list.append(np.column_stack((m_rvs_mean[:, itr], m_rvs_std[:, itr])))

            if prob_dist_list[itr].dist.name == 'truncnorm':
                clip_a, clip_b = Utility.get_standard_normal_clips(*Utility.get_my_normal_clips(*temp),
                                                                   m_rvs_mean[:, itr], m_rvs_std[:, itr])
                out_list.append(np.column_stack((clip_a, clip_b, m_rvs_mean[:, itr], m_rvs_std[:, itr])))

        return out_list, corr_matrix_realizations.transpose((1, 2, 0))

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def fit_dist_to_data(fit_dist, data):
        dist_name = fit_dist.name
        if dist_name == 'lognorm':
            return fit_dist.fit(data, floc=0)
        else:
            return fit_dist.fit(data)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def is_pos_def(x):
        return np.all(np.linalg.eigvals(x) > 0)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_cell_range(sheet, start_col, start_row, end_col, end_row):
        return [[cell.value for cell in sheet.row_slice(row, start_colx=start_col, end_colx=end_col + 1)] for row in
                range(start_row, end_row + 1)]

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def col2index(col):
        num = 0
        for c in col:
            if c in string.ascii_letters:
                num = num * 26 + (ord(c.upper()) - ord('A')) + 1
        return num - 1

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def xlsread(file_path, sheet_name, data_range):
        data_range = data_range.split(':')
        if '.xlsx' in file_path:
            wb = load_workbook(file_path, read_only=True)
            ws = wb[sheet_name]
            # Read the cell values into a list of lists
            data_rows = list()
            for row in ws[data_range[0]:data_range[1]]:
                data_cols = list()
                for cell in row:
                    data_cols.append(cell.value)
                data_rows.append(data_cols)
        elif '.xls' in file_path:
            wb = xlrd.open_workbook(file_path)
            ws = wb.sheet_by_name(sheet_name)
            start_col = ''
            for c in data_range[0]:
                if not Utility.is_numeric_string(c):
                    start_col += c
            start_col = Utility.col2index(start_col)
            start_row = ''
            for c in data_range[0]:
                if Utility.is_numeric_string(c):
                    start_row += c
            start_row = int(start_row) - 1
            end_col = ''
            for c in data_range[1]:
                if not Utility.is_numeric_string(c):
                    end_col += c
            end_col = Utility.col2index(end_col)
            end_row = ''
            for c in data_range[1]:
                if Utility.is_numeric_string(c):
                    end_row += c
            end_row = int(end_row) - 1
            data_rows = Utility.get_cell_range(ws, start_col, start_row, end_col, end_row)
        else:
            return

        return np.array(data_rows)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_exact_mixture_params(mean_req, sig_req, weights):
        vector_size = mean_req.shape[1]
        mean_req_exact = np.sum(
            mean_req * Utility.make_array_2d(weights).T, axis=0)
        var_req_exact = np.sum(
            Utility.make_array_2d(weights).T *
            (np.square(sig_req) + np.square(mean_req - mean_req_exact)),
            axis=0)
        sig_req_exact = np.sqrt(var_req_exact)
        return mean_req_exact.reshape(1, vector_size), sig_req_exact.reshape(1, vector_size)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def calc_oscillator_resp(freq,
                             fourier_amp,
                             osc_damping,
                             osc_freq,
                             max_freq_ratio=5.,
                             peak_resp_only=False,
                             osc_type='psa'):
        """Compute the time series response of an oscillator.
    
        Parameters
        ----------
        freq : array_like
            frequency of the Fourier acceleration spectrum [Hz]
        fourier_amp : array_like
            Fourier acceleration spectrum [g-sec]
        osc_damping : float
            damping of the oscillator [decimal]
        osc_freq : float
            frequency of the oscillator [Hz]
        max_freq_ratio : float, default=5
            minimum required ratio between the oscillator frequency and
            then maximum frequency of the time series. It is recommended that this
            value be 5.
        peak_resp_only : bool, default=False
            If only the peak response is returned.
        osc_type : str, default='psa'
            type of response. Options are:
                'sd': spectral displacement
                'sv': spectral velocity
                'sa': spectral acceleration
                'psv': psuedo-spectral velocity
                'psa': psuedo-spectral acceleration
        Returns
        -------
        response : :class:`numpy.ndarray` or float
            time series response of the oscillator
        
        Copyright 2016-18 Albert Kottke - pyrotd
        
        """
        ang_freq = 2 * np.pi * freq
        osc_ang_freq = 2 * np.pi * osc_freq

        # Single-degree of freedom transfer function
        h = (1 / (ang_freq ** 2. - osc_ang_freq ** 2 -
                  2.j * osc_damping * osc_ang_freq * ang_freq))
        if osc_type == 'sd':
            pass
        elif osc_type == 'sv':
            h *= 1.j * ang_freq
        elif osc_type == 'sa':
            h *= -(ang_freq ** 2)
        elif osc_type == 'psa':
            h *= -(osc_ang_freq ** 2)
        elif osc_type == 'psv':
            h *= -osc_ang_freq
        else:
            raise RuntimeError

        # Adjust the maximum frequency considered. The maximum frequency is 5
        # times the oscillator frequency. This provides that at the oscillator
        # frequency there are at least tenth samples per wavelength.
        n = len(fourier_amp)
        m = max(n, int(max_freq_ratio * osc_freq / freq[1]))
        scale = float(m) / float(n)

        # Scale factor is applied to correct the amplitude of the motion for the
        # change in number of points
        resp = scale * np.fft.irfft(fourier_amp * h, 2 * (m - 1))

        if peak_resp_only:
            resp = np.abs(resp).max()

        return resp

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_response_spectra(gm_data, dt, **kwargs):
        periods = np.array(kwargs.get('periods', np.logspace(np.log10(0.05), np.log10(5), 50, endpoint=True))).flatten()
        xi = kwargs.get('xi', 0.05)

        fourier_amp = np.fft.rfft(gm_data)
        freq = np.linspace(0, 1. / (2 * dt), num=fourier_amp.size)
        spec_accels = [Utility.calc_oscillator_resp(
            freq, fourier_amp, xi, 1 / period, max_freq_ratio=5,
            peak_resp_only=True, osc_type='psa'
        ) for period in periods]

        return np.column_stack((periods, spec_accels))

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def replace_pattern_in_file(file_path, pattern_to_find, pattern_to_replace):
        with open(file_path, 'r') as fid:
            lines = fid.readlines()
        lines = [line.replace(pattern_to_find, pattern_to_replace) for line in lines]
        with open(file_path, 'w') as fid:
            fid.writelines(lines)
        return

    @staticmethod
    def replace_in_files(dir_path, pattern_to_find, pattern_to_replace, ext):
        files = os.listdir(dir_path)
        for file in files:
            if file.split('.')[-1] == ext.split('.')[-1]:
                Utility.replace_pattern_in_file(Utility.get_path(dir_path, file), pattern_to_find, pattern_to_replace)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def mark_max_in_time_hist(ax, time_hist, t, marker, **kwargs):
        markeredgecolor = kwargs.get('markeredgecolor', 'r')
        markersize = kwargs.get('markersize', 10)
        precision = kwargs.get('precision', 5.3)
        units = kwargs.get('units', '')
        textcolor = kwargs.get('textcolor', 'k')
        fontsize = kwargs.get('fontsize', 12)
        ha = kwargs.get('ha', 'left')
        va = kwargs.get('va', 'center')

        ind_max = np.argmax(np.abs(time_hist))
        ax.plot(t[ind_max], time_hist[ind_max], marker, markerfacecolor="None", markeredgecolor=markeredgecolor,
                markersize=markersize)
        ax.text(t[ind_max], time_hist[ind_max], f'{np.abs(time_hist[ind_max]):{precision}} {units}', color=textcolor,
                fontsize=fontsize, horizontalalignment=ha, verticalalignment=va)
        return ax

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def add_common_labels_to_sub_plots(fig, xlabel, ylabel):
        # add a big axis, hide frame
        fig.add_subplot(111, frameon=False)
        ax = fig.gca()
        # hide tick and tick label of the big axis
        ax.tick_params(labelcolor='none', top=False, bottom=False, left=False, right=False)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_xyz_within_lims(x, y, z, x_lim, y_lim, z_lim):
        mask = np.all([x >= x_lim[0], x <= x_lim[-1], y >= y_lim[0], y <= y_lim[-1], z >= z_lim[0], z <= z_lim[-1]],
                      axis=0)
        return x[mask], y[mask], z[mask]

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_xy_within_lims(x, y, x_lim, y_lim):
        mask = np.all([x >= x_lim[0], x <= x_lim[-1], y >= y_lim[0], y <= y_lim[-1]], axis=0)
        return x[mask], y[mask]

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def merge_sorted_1d_arrays(big_array, small_array):
        if np.isscalar(small_array):
            small_array = np.array([small_array])
        big_array = np.array(big_array)
        small_array = np.array(small_array)
        for i_s in range(len(small_array)):
            if not np.any(big_array == small_array[i_s]):
                big_array = np.hstack(
                    (big_array[big_array < small_array[i_s]],
                     small_array[i_s],
                     big_array[big_array > small_array[i_s]]))

        return big_array

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def ecdf(data):
        x = np.sort(data)
        n = x.size
        y = np.arange(1, n + 1) / n
        return x, y

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_total_correlation_estimate(mixture_list, corr_matrix, rng_seed=None):
        # mixture_list : list of marginal mixtures, each with n_s components
        # corr_matrix : (n_rv, n_rv, <n_s>) n_s correlation matrices

        mixture_size = mixture_list[0].mixture_size
        weights = mixture_list[0].weights

        for itr in range(1, len(mixture_list)):
            if np.any(mixture_list[itr].weights != weights):
                return

        if len(corr_matrix.shape) == 2:
            corr_matrix = corr_matrix[:, :, np.newaxis].repeat(mixture_size, axis=2)

        np.random.seed(rng_seed)
        mixture_inds = np.random.choice(mixture_size, size=500, p=weights)
        rvs_temp = np.array([]).reshape(0, len(mixture_list))
        for itr in range(len(mixture_inds)):
            ind = mixture_inds[itr]
            prob_dist_list = [item.prob_dists[ind] if item.each == 'arbitrary'
                              else item.each(*np.array(item.args)[ind, :]) for item in mixture_list]
            m = multivariate_nataf(prob_dist_list, corr_matrix[:, :, ind])
            if rng_seed is not None:
                rng_seed_send = Utility.get_num([rng_seed, itr])
            else:
                rng_seed_send = rng_seed
            rvs_temp = np.vstack((rvs_temp,
                                  m.rvs(size=1, random_state=rng_seed_send, method='mcs')
                                  ))

        return np.corrcoef(rvs_temp, rowvar=False)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def plot_2d_sampling_steps(prob_dist_list, corr_matrix, sample_size, sampling_method, **kwargs):
        rng_seed = kwargs.get('rng_seed', None)
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        k = kwargs.get('k', 5)
        n_cont = kwargs.get('n_cont', 100)
        ms = kwargs.get('ms', 10)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1)
        mc = kwargs.get('mc', 'red')
        n_bins = kwargs.get('n_bins', 10)
        hc = kwargs.get('hc', 'gray')
        same_lim = kwargs.get('same_lim', False)
        dist_func = kwargs.get('dist_func', 'pdf')
        fig_ax = kwargs.get('fig_ax', None)
        scilimits = kwargs.get('scilimits', (-4, 4))

        num = figkwargs.get('num', None)
        figkwargs['num'] = num

        m = multivariate_nataf(prob_dist_list, corr_matrix)
        x, independent_samples, dependent_samples, uniform_dependent_samples, transformed_samples = m.sampler(
            sample_size, rng_seed, sampling_method, return_all=True
        )

        if sampling_method == 'lhs':
            grid, ind = m.get_2d_lhs_grid(sample_size, n_cont)
            x_grid, u_grid, z_grid, x_star_grid, y_grid = grid
        else:
            ind, x_grid, u_grid, z_grid, x_star_grid, y_grid = (None,) * 6

        export_mat_dict = dict()
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig_1 = plt.figure(**figkwargs)
            ax_1 = fig_1.gca()
        else:
            fig_1 = fig_ax[0][0]
            ax_1 = fig_ax[1][0]
        ax_1.minorticks_on()
        ax_1.grid(True, which="major", alpha=grid_alpha)
        ax_1.grid(True, which="minor", alpha=minor_grid_alpha)
        ax_1.plot(x[:, 0], x[:, 1], '.', markersize=ms, color=mc)
        ax_1.ticklabel_format(axis='both', scilimits=scilimits)
        if sampling_method == 'lhs':
            ax_1.grid(False, which="both")
            ax_1 = Utility.plot_grid(ax_1, x_grid[:, :, 0], x_grid[:, :, 1], ind, lw, lc, grid_alpha)
            export_mat_dict['ind'] = ind
            export_mat_dict['x_grid'] = x_grid
        ax_1.set_xlim((0, 1))
        ax_1.set_ylim((0, 1))
        export_mat_dict['x'] = x

        figkwargs['num'] = fig_1.number + 1
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig_ax_2 = None
        else:
            fig_ax_2 = [fig_ax[0][1], fig_ax[1][1]]
        fig_2, ax_2, temp_dict = Utility.scatter_hist_2d(independent_samples,
                                                         multivariate_nataf([norm(), norm()], np.eye(2),
                                                                            corr_matrix_z=np.eye(2)),
                                                         figkwargs=figkwargs,
                                                         grid_alpha=grid_alpha,
                                                         minor_grid_alpha=minor_grid_alpha,
                                                         k=k, n_cont=n_cont,
                                                         n_bins=n_bins, ms=ms, mc=mc,
                                                         lw=lw, lc=lc, hc=hc,
                                                         same_lim=same_lim, dist_func=dist_func, fig_ax=fig_ax_2)
        ax_2[0].ticklabel_format(axis='both', scilimits=scilimits)
        if sampling_method == 'lhs':
            ax_2[0].grid(False, which="both")
            ax_2[0] = Utility.plot_grid(ax_2[0], u_grid[:, :, 0], u_grid[:, :, 1], ind, lw, lc, grid_alpha)
            export_mat_dict['u_grid'] = u_grid
        export_mat_dict['u'] = independent_samples
        export_mat_dict[f'u_{dist_func}_plot'] = np.dstack((temp_dict[f'{dist_func}_plot_1'],
                                                            temp_dict[f'{dist_func}_plot_2']))
        export_mat_dict['u_joint_pdf_surf'] = temp_dict[f'joint_pdf_surf']

        figkwargs['num'] = fig_2.number + 1
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig_ax_3 = None
        else:
            fig_ax_3 = [fig_ax[0][2], fig_ax[1][2]]
        fig_3, ax_3, temp_dict = Utility.scatter_hist_2d(dependent_samples,
                                                         multivariate_nataf([norm(), norm()], m.corr_matrix_z,
                                                                            corr_matrix_z=m.corr_matrix_z),
                                                         figkwargs=figkwargs,
                                                         grid_alpha=grid_alpha,
                                                         minor_grid_alpha=minor_grid_alpha,
                                                         k=k, n_cont=n_cont,
                                                         n_bins=n_bins, ms=ms, mc=mc,
                                                         lw=lw, lc=lc, hc=hc,
                                                         same_lim=same_lim, dist_func=dist_func, fig_ax=fig_ax_3)
        ax_3[0].ticklabel_format(axis='both', scilimits=scilimits)
        if sampling_method == 'lhs':
            ax_3[0].grid(False, which="both")
            ax_3[0] = Utility.plot_grid(ax_3[0], z_grid[:, :, 0], z_grid[:, :, 1], ind, lw, lc, grid_alpha)
            export_mat_dict['z_grid'] = z_grid
        export_mat_dict['z'] = dependent_samples
        export_mat_dict[f'z_{dist_func}_plot'] = np.dstack((temp_dict[f'{dist_func}_plot_1'],
                                                            temp_dict[f'{dist_func}_plot_2']))
        export_mat_dict['z_joint_pdf_surf'] = temp_dict[f'joint_pdf_surf']
        export_mat_dict['z_corr_matrix'] = m.corr_matrix_z

        figkwargs['num'] = fig_3.number + 1
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig_4 = plt.figure(**figkwargs)
            ax_4 = fig_4.gca()
        else:
            fig_4 = fig_ax[0][3]
            ax_4 = fig_ax[1][3]
        ax_4.minorticks_on()
        ax_4.grid(True, which="major", alpha=grid_alpha)
        ax_4.grid(True, which="minor", alpha=minor_grid_alpha)
        ax_4.ticklabel_format(axis='both', scilimits=scilimits)
        ax_4.plot(uniform_dependent_samples[:, 0], uniform_dependent_samples[:, 1], '.', markersize=ms, color=mc)
        if sampling_method == 'lhs':
            ax_4.grid(False, which="both")
            ax_4 = Utility.plot_grid(ax_4, x_star_grid[:, :, 0], x_star_grid[:, :, 1], ind, lw, lc, grid_alpha)
            export_mat_dict['x*_grid'] = x_star_grid
        ax_4.set_xlim((0, 1))
        ax_4.set_ylim((0, 1))
        export_mat_dict['x*'] = uniform_dependent_samples

        figkwargs['num'] = fig_4.number + 1
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig_ax_5 = None
        else:
            fig_ax_5 = [fig_ax[0][4], fig_ax[1][4]]
        fig_5, ax_5, temp_dict = Utility.scatter_hist_2d(transformed_samples, m,
                                                         figkwargs=figkwargs,
                                                         grid_alpha=grid_alpha,
                                                         minor_grid_alpha=minor_grid_alpha,
                                                         k=k, n_cont=n_cont,
                                                         n_bins=n_bins, ms=ms, mc=mc,
                                                         lw=lw, lc=lc, hc=hc,
                                                         same_lim=same_lim, dist_func=dist_func, fig_ax=fig_ax_5)
        ax_5[0].ticklabel_format(axis='both', scilimits=scilimits)
        if sampling_method == 'lhs':
            ax_5[0].grid(False, which="both")
            ax_5[0] = Utility.plot_grid(ax_5[0], y_grid[:, :, 0], y_grid[:, :, 1], ind, lw, lc, grid_alpha)
            export_mat_dict['y_grid'] = y_grid
        export_mat_dict['y'] = transformed_samples
        export_mat_dict[f'y_{dist_func}_plot'] = np.dstack((temp_dict[f'{dist_func}_plot_1'],
                                                            temp_dict[f'{dist_func}_plot_2']))
        export_mat_dict['y_joint_pdf_surf'] = temp_dict[f'joint_pdf_surf']

        return [fig_1, fig_2, fig_3, fig_4, fig_5], [ax_1, ax_2, ax_3, ax_4, ax_5], export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def plot_grid(ax, xx, yy, ind, lw, lc, grid_alpha):
        for itr in range(1, len(ind) - 1):
            ax.plot(xx[:, ind[itr]], yy[:, ind[itr]], '-', linewidth=lw, color=lc, alpha=grid_alpha, zorder=1)
            ax.plot(xx[ind[itr], :], yy[ind[itr], :], '-', linewidth=lw, color=lc, alpha=grid_alpha, zorder=1)
        return ax

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def scatter_hist_2d(data, mvnataf, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        k = kwargs.get('k', 5)
        n_cont = kwargs.get('n_cont', 100)
        n_bins = kwargs.get('n_bins', 10)
        ms = kwargs.get('ms', 10)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1)
        hc = kwargs.get('hc', 'gray')
        mc = kwargs.get('mc', 'red')
        same_lim = kwargs.get('same_lim', False)
        dist_func = kwargs.get('dist_func', 'pdf')
        fig_ax = kwargs.get('fig_ax', None)

        left, width = 0.1, 0.65
        bottom, height = 0.1, 0.65
        spacing = 0.005

        rect_scatter = [left, bottom, width, height]
        rect_histx = [left, bottom + height + spacing, width, 0.2]
        rect_histy = [left + width + spacing, bottom, 0.2, height]

        # start with a rectangular Figure
        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            axs = list()
            axs.append(fig.add_axes(rect_scatter))
            axs.append(fig.add_axes(rect_histx))
            axs.append(fig.add_axes(rect_histy))
        else:
            fig = fig_ax[0]
            axs = fig_ax[1]

        axs[0].minorticks_on()
        axs[0].grid(True, which="major", alpha=grid_alpha)
        axs[0].grid(True, which="minor", alpha=minor_grid_alpha)
        axs[0].tick_params(direction='in', top=True, right=True)
        axs[1].tick_params(direction='in', labelbottom=False)
        axs[2].tick_params(direction='in', labelleft=False)

        # the scatter plot:
        x = data[:, 0]
        y = data[:, 1]
        axs[0].plot(x, y, '.', color=mc, markersize=ms)
        if dist_func == 'cdf':
            cumulative = True
        else:
            cumulative = False

        axs[1].hist(x, bins=n_bins, density=True, color=hc, edgecolor='white', cumulative=cumulative)
        axs[2].hist(y, bins=n_bins, orientation='horizontal', density=True, color=hc, edgecolor='white',
                    cumulative=cumulative)

        export_mat_dict = dict()
        prob_dist_list = mvnataf.prob_dist_list
        for itr in range(len(prob_dist_list)):
            ax = axs[itr + 1]
            prob_dist = prob_dist_list[itr]
            mean = prob_dist.mean()
            std = prob_dist.std()
            x1 = mean - k * std
            x2 = mean + k * std
            x = np.linspace(x1, x2, n_cont)
            if dist_func == 'cdf':
                df_x = prob_dist.cdf(x)
            else:
                df_x = prob_dist.pdf(x)
            if itr == 0:
                ax.plot(x, df_x, color=lc, linewidth=lw)
            if itr == 1:
                ax.plot(df_x, x, color=lc, linewidth=lw)

            ax.axis('off')
            export_mat_dict[f'{dist_func}_plot_{itr + 1}'] = np.column_stack((x, df_x))
            export_mat_dict[f'dist_name_{itr + 1}'] = prob_dist.dist.name

        export_mat_dict[f'data'] = data

        mu_1 = prob_dist_list[0].mean()
        mu_2 = prob_dist_list[1].mean()
        sig_1 = prob_dist_list[0].std()
        sig_2 = prob_dist_list[1].std()
        xlim = (mu_1 - k * sig_1, mu_1 + k * sig_1)
        ylim = (mu_2 - k * sig_2, mu_2 + k * sig_2)
        x_bounds = prob_dist_list[0].support()
        y_bounds = prob_dist_list[1].support()
        xlim = (max(xlim[0], x_bounds[0]), min(xlim[1], x_bounds[1]))
        ylim = (max(ylim[0], y_bounds[0]), min(ylim[1], y_bounds[1]))

        x_surf = np.linspace(*xlim, n_cont)
        y_surf = np.linspace(*ylim, n_cont)
        x_surf = x_surf[1:]
        y_surf = y_surf[1:]
        surf_xx, surf_yy = np.meshgrid(x_surf, y_surf)
        zz = mvnataf.pdf(np.dstack((surf_xx, surf_yy)))
        axs[0].contour(surf_xx, surf_yy, zz)
        export_mat_dict[f'joint_pdf_surf'] = np.dstack([surf_xx, surf_yy, zz])

        if same_lim:
            xlim = (min(xlim[0], ylim[0]), max(xlim[1], ylim[1]))
            ylim = xlim
        axs = Utility.adjust_scatter_hist_2d_axes(axs, xlim, ylim)

        return fig, axs, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def adjust_scatter_hist_2d_axes(axs, xlim, ylim):
        axs[0].set_xlim(xlim)
        axs[0].set_ylim(ylim)
        axs[1].set_xlim(axs[0].get_xlim())
        axs[2].set_ylim(axs[0].get_ylim())
        return axs

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def plot_3d_bar(x, y, dz, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        cmap = kwargs.get('cmap', 'viridis')
        minor_grid = kwargs.get('minor_grid', False)
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)  # also accepts tight_layout=bool
            ax = fig.add_subplot(projection='3d')
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        if minor_grid:
            ax.minorticks_on()
        else:
            ax.minorticks_off()

        temp = np.diff(x)
        dx = np.ones(len(x)) * min(temp[temp > 0])

        temp = np.diff(y)
        dy = np.ones(len(y)) * min(temp[temp > 0])

        fracs = dz / dz.max()
        normalize = colors.Normalize(fracs.min(), fracs.max())
        my_cmap = cm.get_cmap(cmap)
        color_values = my_cmap(normalize(fracs.tolist()))

        z = np.zeros(len(x))

        ax.bar3d(x, y, z, dx, dy, dz, color=color_values)
        ax_in = inset_axes(ax, width="50%", height="5%", loc='upper right')

        sm = cm.ScalarMappable(cmap=my_cmap)
        sm.set_array(dz)
        fig.colorbar(sm, cax=ax_in, orientation='horizontal')

        export_mat_dict = dict()
        export_mat_dict['x'] = x
        export_mat_dict['y'] = y
        export_mat_dict['dz'] = dz

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def scatter_hist_1d(data, prob_dist=None, **kwargs):
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        k = kwargs.get('k', 5)
        n_cont = kwargs.get('n_cont', 100)
        n_bins = kwargs.get('n_bins', 10)
        ms = kwargs.get('ms', 10)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1)
        hc = kwargs.get('hc', 'gray')
        mc = kwargs.get('mc', 'red')
        dist_func = kwargs.get('dist_func', 'pdf')
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        if dist_func == 'cdf':
            cumulative = True
        else:
            cumulative = False
        bin_hgts, bin_edges, _ = ax.hist(data, bins=n_bins, density=True, color=hc, edgecolor='white',
                                         cumulative=cumulative)
        _, _, patches = ax.hist(data, bins=n_bins, density=True, color=None, edgecolor=None, cumulative=cumulative,
                                histtype='step', linewidth=0)
        ax.plot(data, 0 * data, '.', color=mc, markersize=ms, zorder=10, clip_on=False)

        export_mat_dict = dict()

        if prob_dist is not None:
            mean = prob_dist.mean()
            std = prob_dist.std()
            x1 = mean - k * std
            x2 = mean + k * std
            x = np.linspace(x1, x2, n_cont)
            if dist_func == 'cdf':
                df_x = prob_dist.cdf(x)
            else:
                df_x = prob_dist.pdf(x)
            ax.plot(x, df_x, color=lc, linewidth=lw)

            xlim = (x1, x2)
            x_bounds = prob_dist.support()
            xlim = (max(xlim[0], x_bounds[0]), min(xlim[1], x_bounds[1]))
            ax.set_xlim(xlim)

            export_mat_dict[f'{dist_func}_plot'] = np.column_stack((x, df_x))
            export_mat_dict['dist_name'] = prob_dist.dist.name

        export_mat_dict['data'] = data
        export_mat_dict['patches'] = patches
        export_mat_dict['bin_hgts'] = bin_hgts
        export_mat_dict['bin_edges'] = bin_edges

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_rv_x_lims(pd, k=10):
        mean = pd.mean()
        std = pd.std()
        supp = pd.support()
        return max(mean - k * std, supp[0]), min(mean + k * std, supp[-1])

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def shell_exec(exec_file_name, cd_path, exec_commands, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        local_bash_path = kwargs.get('local_bash_path', '')
        remove_file = kwargs.get('remove_file', True)

        write_lines = list()
        write_lines.append('#!/bin/bash\n')
        write_lines.append('cd ' + '"' + cd_path + '"\n')
        for command in exec_commands:
            write_lines.append(command + '\n')
        write_lines.append('cd ' + '"' + Utility.get_path(os.getcwd()) + '"\n')

        with open(exec_file_name, 'w') as fid:
            fid.writelines(write_lines)

        if comp_env == 'local' and platform.system() == 'Windows':
            subprocess.call(f"\"{local_bash_path}\" \"{exec_file_name}\"", shell=True)
        else:
            os.system(f'chmod +x "{exec_file_name}"')
            os.system(f'./{exec_file_name}')

        if remove_file:
            os.remove(exec_file_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_curr_time_stamp():
        curr_date_time = datetime.now()
        temp = str(curr_date_time).split()
        curr_date = temp[0]
        curr_time = '-'.join([str(i) for i in [int(float(t)) for t in temp[1].split(':')]])
        return f'date_{curr_date}_time_{curr_time}'

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def untar_file(tar_file_path, **kwargs):

        local_bash_path = kwargs.get('local_bash_path', '')
        comp_env = kwargs.get('comp_env', 'local')
        flags = kwargs.get('flags', ['x', 'z', 'v', 'f'])

        tar_string = f"tar -{''.join(flags)} {os.path.basename(tar_file_path)}"

        exec_commands = [tar_string]
        Utility.shell_exec('UNTAR', os.path.dirname(tar_file_path), exec_commands,
                           comp_env=comp_env, local_bash_path=local_bash_path)

        return

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_rayleigh_damping_params(xi_1, w_1, xi_2, w_2):
        denom = w_1 ** 2 - w_2 ** 2
        alpha = 2. * w_1 * w_2 * (xi_2 * w_1 - xi_1 * w_2) / denom
        beta = (2. * xi_1 * w_1 - 2. * xi_2 * w_2) / denom
        return alpha, beta

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def plot_log_grid(a, axis, **kwargs):
        major_grid_alpha = kwargs.get('major_grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.5)
        major_lw = kwargs.get('major_lw', 1)
        minor_lw = kwargs.get('minor_lw', 1)
        major_color = kwargs.get('major_color', 'lightgrey')
        minor_color = kwargs.get('minor_color', 'lightgrey')

        if axis.lower() == 'x':
            axis_lim_power = np.rint(np.log10(np.array(a.get_xlim())))
            other_lim = a.get_ylim()
        elif axis.lower() == 'y':
            axis_lim_power = np.rint(np.log10(np.array(a.get_ylim())))
            other_lim = a.get_xlim()
        else:
            raise ValueError("Invalid axis!")

        major_grid = 10. ** np.arange(axis_lim_power[0], axis_lim_power[1] + 1)
        minor_grid = []
        for i in range(len(major_grid) - 1):
            minor_grid += list(np.linspace(major_grid[i], major_grid[i + 1], 10))
        minor_grid = np.unique(np.array(minor_grid))
        minor_grid = np.array([item for item in minor_grid if item not in major_grid])

        if axis.lower() == 'x':
            for i in range(len(major_grid)):
                a.plot([major_grid[i], major_grid[i]], [other_lim[0], other_lim[1]],
                       lw=major_lw, color=major_color, alpha=major_grid_alpha, zorder=0)
            for i in range(len(minor_grid)):
                a.plot([minor_grid[i], minor_grid[i]], [other_lim[0], other_lim[1]],
                       lw=minor_lw, color=minor_color, alpha=minor_grid_alpha, zorder=0)
        elif axis.lower() == 'y':
            for i in range(len(major_grid)):
                a.plot([other_lim[0], other_lim[1]], [major_grid[i], major_grid[i]],
                       lw=major_lw, color=major_color, alpha=major_grid_alpha, zorder=0)
            for i in range(len(minor_grid)):
                a.plot([other_lim[0], other_lim[1]], [minor_grid[i], minor_grid[i]],
                       lw=minor_lw, color=minor_color, alpha=minor_grid_alpha, zorder=0)

        return a

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_many_dists(dist, params):
        # params must be a 2-d array
        return dist(*params.T)

    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def get_ylim(ax, thresh):
        y_lim_0 = np.inf
        y_lim_1 = -np.inf

        for line in ax.lines:
            y_data = line.get_data()[1]
            if y_data.size == 0:
                continue  # Skip lines with no data
            curr_min = np.min(y_data)
            curr_max = np.max(y_data)
            if curr_min < y_lim_0:
                y_lim_0 = curr_min
            if curr_max > y_lim_1:
                y_lim_1 = curr_max

        # If no valid data was found, return current ylim
        if not np.isfinite(y_lim_0) or not np.isfinite(y_lim_1):
            return ax.get_ylim()

        # Apply thresholds
        ylim = ax.get_ylim()
        if thresh[0] is not None and thresh[1] is None:
            ylim = (y_lim_0 * thresh[0], ylim[1])
        elif thresh[0] is None and thresh[1] is not None:
            ylim = (ylim[0], y_lim_1 * thresh[1])
        elif thresh[0] is not None and thresh[1] is not None:
            ylim = (y_lim_0 * thresh[0], y_lim_1 * thresh[1])

        return ylim



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 16 12:42:26 2019

@author: Angshuman Deb
"""

from .matplotlib_setup import *
from .analysis import Analysis
from .avg_sa import AvgSa
from .ds import DS
from .edp import EDP, FrameMaxDeformation, MaxColRebarStrain, MaxSpringDeformation
from .gms import GMS
from .im import IM
from .interp_exterp_model import InterpExterpModel
from .mixture import mixture
from .multivariate_nataf import multivariate_nataf
from .nltha import NLTHA
from .prelim_analysis import PrelimAnalysis
from .psdamha import PSDamHA
from .psdemha import PSDemHA
from .psha import PSHA
from .sa import Sa
from .sa_t import SaT
from .structural_analysis_platform import StructuralAnalysisPlatform, OpenSeesTcl, OpenSeesPy
from .structure import Structure, FrameStructure, OSB
from .utility import Utility


__all__ = [
    'Analysis',
    'AvgSa',
    'DS',
    'EDP', 'FrameMaxDeformation', 'MaxColRebarStrain', 'MaxSpringDeformation',
    'GMS',
    'IM',
    'InterpExterpModel',
    'mixture',
    'multivariate_nataf',
    'NLTHA',
    'PrelimAnalysis',
    'PSDamHA',
    'PSDemHA',
    'PSHA',
    'Sa',
    'SaT',
    'StructuralAnalysisPlatform', 'OpenSeesTcl', 'OpenSeesPy',
    'Structure', 'FrameStructure', 'OSB',
    'Utility',
]

__author__ = 'Angshuman Deb'
__title__ = 'pypbee'
__version__ = "0.1.0.post1"



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  8 16:31:34 2019

@author: Angshuman Deb
"""

from .analysis import Analysis
from .utility import Utility
from .structure import Structure
import os
import numpy as np
from time import time
from pathos.multiprocessing import ProcessPool as Pool
from functools import partial
import scipy.io as sio
from benedict import benedict
import matplotlib.pyplot as plt
import matplotlib.colors as colors


class PSDamHA(Analysis):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, ds_list, im=None, **kwargs):
        comp_env = kwargs.get('comp_env', 'local')
        self.sol_type = kwargs.get('sol_type', 'numerical')
        self.ds_list = ds_list
        self.im = im
        super().__init__(self.__class__.__name__, ds_list[0].structure, comp_env)

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def run_parallel(setup_dir_path, compute_damage_hazard_integral_list, sol_type,
                     im, haz_lev_list, n_gm_list, delta_input_list, min_max_scale_fac_list,
                     node_index, batch_index, analysis_index, run_case):

        start_time = time()

        for_which = list(run_case.astype(str))
        analysis_case = for_which[Structure.get_dir_level_index('Analysis_Case')]
        run_case_str = ' '.join(for_which)

        psdamha_results = list()
        for itr in range(len(compute_damage_hazard_integral_list)):
            temp = compute_damage_hazard_integral_list[itr](for_which, im, haz_lev_list, n_gm_list,
                                                            min_max_scale_fac=min_max_scale_fac_list[itr],
                                                            delta_input=delta_input_list[itr],
                                                            sol_type=sol_type)
            psdamha_results.append(temp)

        status_file_name = f'status_analysis_index_{int(analysis_index)}' \
                           f'_node_{int(node_index)}_batch_{int(batch_index)}.txt'
        status_file_path = Utility.get_path(setup_dir_path, analysis_case, status_file_name)

        with open(status_file_path, 'w') as fid:
            fid.write(f"{run_case_str} : COMPLETED in {time() - start_time} seconds\n")

        return f'{run_case_str}', psdamha_results

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def stage(self, analysis_case, design_num_list, rng_seed):

        ds_list = self.ds_list

        results_dir_path = self.get_results_dir_path()

        set_rng_seed, seed_multiplier = Utility.setup_rng(rng_seed)

        for design_num in design_num_list:

            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)

            if not os.path.isdir(target_work_dir_path):
                os.makedirs(target_work_dir_path)

            normalized_fragility_dist_param_vals = dict()

            for ds in ds_list:
                if set_rng_seed:
                    rng_seed_send = seed_multiplier * Utility.get_num([int(design_num), ds.tag])
                else:
                    rng_seed_send = rng_seed

                normalized_fragility_dist_param_vals[ds.tag] = \
                    ds.generate_normalized_fragility_dist_param_vals(analysis_case, rng_seed=rng_seed_send)

            normalized_fragility_dist_param_vals_file_path = Utility.get_path(
                target_work_dir_path, 'normalized_fragility_dist_param_vals.pickle')
            Utility.pickle_dump_dict(normalized_fragility_dist_param_vals_file_path,
                                     normalized_fragility_dist_param_vals)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def setup(self, analysis_case, design_num_list, haz_lev_list=None, n_gm_list=None, **kwargs):
        rng_seed = kwargs.get('rng_seed', None)
        run_time_limit = kwargs.get('run_time_limit', '01:00:00')
        allocation_name = kwargs.get('allocation_name', '')
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        pool_size_factor = kwargs.get('pool_size_factor', 1)

        structure = self.structure

        self.stage(analysis_case, design_num_list, rng_seed)

        setup_dir_path = self.get_setup_dir_path()
        if not os.path.isdir(Utility.get_path(setup_dir_path, analysis_case)):
            os.makedirs(Utility.get_path(setup_dir_path, analysis_case))

        temp = list()
        for design_num in design_num_list:
            temp.append(
                Utility.get_multi_level_iterator_cases(
                    int(analysis_case),
                    int(design_num),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[2] + 1)),
                    list(range(1, structure.get_random_model_param_vals([analysis_case, design_num]).shape[0] + 1)),
                )
            )
        analysis_list = np.vstack(temp)

        list_file_path = Utility.get_path(setup_dir_path, analysis_case, 'analysis_list.txt')
        Utility.save_array_as_text_file(list_file_path, analysis_list, fmt='%d',
                                        header=f'Total analyses to run = {analysis_list.shape[0]:d}')

        self.set_analysis_setup_info(
            analysis_case,
            {'design_num_list': design_num_list,
             'haz_lev_list': haz_lev_list,
             'n_gm_list': n_gm_list,
             }
        )

        if not self.comp_env == 'local':
            self.setup_non_local_run(analysis_case, n_batch, n_job, pool_size_factor, run_time_limit, allocation_name)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def run(self, analysis_case, pool_size, **kwargs):
        ds_list = self.ds_list
        n_batch = kwargs.get('n_batch', 1)
        n_job = kwargs.get('n_job', 1)
        node_index = kwargs.get('node_index', 1)
        batch_index = kwargs.get('batch_index', 1)
        delta_input_list = kwargs.get('delta_input_list', [np.array([])] * len(ds_list))
        min_max_scale_fac_list = kwargs.get('min_max_scale_fac_list', [[1, 1]] * len(ds_list))

        compute_damage_hazard_integral_list = [ds.compute_damage_hazard_integral for ds in ds_list]
        im = self.im

        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()
        analysis_list, analysis_range, n_sim = self.get_analysis_info(analysis_case, 'analysis_list.txt',
                                                                      n_batch, n_job, node_index, batch_index)

        analysis_setup_info = self.get_analysis_setup_info(analysis_case)
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))
        haz_lev_list = analysis_setup_info['haz_lev_list']
        n_gm_list = analysis_setup_info['n_gm_list']

        to_run = partial(PSDamHA.run_parallel, setup_dir_path, compute_damage_hazard_integral_list, self.sol_type,
                         im, haz_lev_list, n_gm_list, delta_input_list, min_max_scale_fac_list,
                         node_index, batch_index)

        to_pass = [[analysis_index for analysis_index in analysis_range],
                   [analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        if pool_size > 1:
            my_pool = Pool(pool_size)
            psdamha_results = dict(my_pool.map(to_run, *to_pass))
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            psdamha_results = dict(list(map(to_run, *to_pass)))

        ds_tags = [ds.tag for ds in ds_list]

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            psdamha_results_design_num = {key: psdamha_results[key] for key in psdamha_results.keys() if
                                          key.split()[Structure.get_dir_level_index('Design_Num')] == design_num}
            psdamha_results_ds = {
                ds_tag: {analysis_key: psdamha_results_design_num[analysis_key][ds_tags.index(ds_tag)]
                         for analysis_key in psdamha_results_design_num.keys()}
                for ds_tag in ds_tags
            }

            if n_batch == 1 and n_job == 1:
                Utility.pickle_dump_dict(Utility.get_path(target_work_dir_path, f'psdamha_results.pickle'),
                                         psdamha_results_ds)
            else:
                Utility.pickle_dump_dict(
                    Utility.get_path(
                        target_work_dir_path, f'psdamha_results_node_{node_index}_batch_{batch_index}.pickle'
                    ), psdamha_results_ds
                )

        # Write final_status_node_{}_batch_{}.txt
        contains_str = f'_node_{int(node_index)}_batch_{int(batch_index)}'
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          [f'contains status_analysis_index_ {contains_str}'],
                                          f'final_status{contains_str}.txt',
                                          file_ext='.txt', delete_original=True)

        return

    # ------------------------------------------------------------------------------------------------------------------

    def wrap_up(self, analysis_case):
        setup_dir_path = self.get_setup_dir_path()
        results_dir_path = self.get_results_dir_path()

        # Write final_status.txt
        Utility.combine_contents_of_files(Utility.get_path(setup_dir_path, analysis_case),
                                          ['startswith final_status_', 'startswith status_analysis_index_'],
                                          'final_status.txt',
                                          file_ext='.txt', delete_original=True)

        analysis_list, _, _ = self.get_analysis_info(analysis_case, 'analysis_list.txt', 1, 1, 1, 1)

        # Combine results from all nodes and batches
        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path, analysis_case, design_num)
            files = os.listdir(target_work_dir_path)
            files = [file for file in files if file.startswith('psdamha_results') and file.endswith(
                '.pickle') and 'node' in file and 'batch' in file]
            if len(files) > 0:
                list_of_dicts = [Utility.pickle_load_dict(Utility.get_path(target_work_dir_path, file)) for file in
                                 files]

                psdamha_results_file_path = Utility.get_path(target_work_dir_path, 'psdamha_results.pickle')
                if os.path.isfile(psdamha_results_file_path):
                    psdamha_results = benedict(Utility.pickle_load_dict(psdamha_results_file_path))
                else:
                    psdamha_results = benedict()

                for d in list_of_dicts:
                    psdamha_results.merge(d)

                Utility.pickle_dump_dict(psdamha_results_file_path, psdamha_results)

                for file in files:
                    os.remove(Utility.get_path(target_work_dir_path, file))

        return

    # ------------------------------------------------------------------------------------------------------------------

    def get_d_star(self, analysis_case, target_mrp_list, d_list, **kwargs):
        exclude_ds_tags = kwargs.get('exclude_ds_tags', list())
        ds_list = self.ds_list
        ds_list = [ds for ds in ds_list if ds.tag not in exclude_ds_tags]

        x_star_list = []
        d_star_list = []
        for (ds, tgt) in zip(ds_list, target_mrp_list):
            d_star_temp, x_star_temp = ds.get_d_star(analysis_case, d_list, tgt)
            x_star_list.append(x_star_temp)
            d_star_list.append(d_star_temp)
        x_star_list = np.array(x_star_list)
        d_star_list = np.array(d_star_list)

        x_star = np.max(x_star_list)
        ind = np.argmax(x_star_list)
        d_star = d_star_list[ind, :]

        return d_star, x_star

    # ------------------------------------------------------------------------------------------------------------------

    def plot_feasible_domain(self, analysis_case, target_mrp_list, color_list, **kwargs):
        mesh_size = kwargs.get('mesh_size', 100)
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        figkwargs = kwargs.get('figkwargs', dict())
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        patch_alpha = kwargs.get('patch_alpha', 0.25)
        lw = kwargs.get('lw', 2)
        mc_gp = kwargs.get('mc_gp', 'red')
        ms_gp = kwargs.get('ms_gp', 6)
        mt_gp = kwargs.get('mt_gp', 'o')
        mc_ngp = kwargs.get('mc_ngp', 'red')
        ms_ngp = kwargs.get('ms_ngp', 6)
        mt_ngp = kwargs.get('mt_ngp', 'd')
        mc_ad = kwargs.get('mc_ad', 'red')
        ms_ad = kwargs.get('ms_ad', 18)
        mt_ad = kwargs.get('mt_ad', '*')
        mc_mps = kwargs.get('mc_mps', list())
        ms_mps = kwargs.get('ms_mps', list())
        mt_mps = kwargs.get('mt_mps', list())
        mark_points = kwargs.get('mark_points', list())
        exclude_ds_tags = kwargs.get('exclude_ds_tags', list())
        fig_ax = kwargs.get('fig_ax', None)
        domain_type = kwargs.get('domain_type', 'full_fledged')
        d_list = kwargs.get('d_list', [])
        da_list = kwargs.get('da_list', [])

        if domain_type not in ['full_fledged', 'simplified_bilinear']:
            raise ValueError("Invalid domain_type!")

        surf_type = 'piecewise_linear'
        if domain_type == 'simplified_bilinear':
            if len(d_list) < 2:
                raise ValueError("Invalid d_list!")
            surf_type = 'bilinear_contours'
            if len(da_list) != 2:
                raise ValueError("Invalid da_list!")

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        ds_list = self.ds_list
        ds_list = [ds for ds in ds_list if ds.tag not in exclude_ds_tags]
        structure = self.structure
        model_params = structure.model_params
        design_vector = np.array(list(model_params['primary_design_params']['value_list_dict'].values()))
        design_num_list = list(model_params['primary_design_params']['value_list_dict'].keys())

        export_mat_dict = dict()
        export_mat_dict['design_vector'] = np.array(design_vector)
        export_mat_dict['design_num_list'] = np.array(design_num_list)
        export_mat_dict['target_mrp_list'] = np.array(target_mrp_list).astype(float)
        export_mat_dict['mesh_size'] = mesh_size

        safe_zone = np.zeros((mesh_size, mesh_size))

        for itr in range(len(ds_list)):
            ds = ds_list[itr]
            color = color_list[itr]
            target_mrp = target_mrp_list[itr]
            data, surface = ds.get_damage_mrp_interp_surface(analysis_case, mesh_size, surf_type,
                                                             d_list=d_list, da_list=da_list)
            ax.contour(surface[:, :, 0], surface[:, :, 1], surface[:, :, -1], [target_mrp],
                       colors=color, linewidths=lw)
            safe_zone[surface[:, :, -1] < target_mrp] = 1
            tag = ds.tag
            export_mat_dict[f'mrp_ds_{tag}'] = data[:, -1]
            export_mat_dict[f'surface_ds_{tag}'] = surface

        cmap = colors.ListedColormap([[0, 1, 0, 1], [1, 1, 1, 1]])

        ax.imshow(safe_zone,
                  extent=[min(design_vector[:, 0]),
                          max(design_vector[:, 0]),
                          min(design_vector[:, 1]),
                          max(design_vector[:, 1])],
                  aspect='auto',
                  origin='lower',
                  alpha=patch_alpha,
                  cmap=cmap)

        fig, ax, _ = structure.plot_design_space(
            grid_alpha=grid_alpha, minor_grid_alpha=minor_grid_alpha,
            save_mat=False, lw=0,
            mc_gp=mc_gp, ms_gp=ms_gp, mt_gp=mt_gp,
            mc_ngp=mc_ngp, ms_ngp=ms_ngp, mt_ngp=mt_ngp,
            mc_ad=mc_ad, ms_ad=ms_ad, mt_ad=mt_ad,
            mc_mps=mc_mps, ms_mps=ms_mps, mt_mps=mt_mps, mark_points=mark_points,
            fig_ax=[fig, ax]
        )

        export_mat_dict['safe_zone'] = safe_zone

        if save_mat:
            name = structure.name
            file_name = f'plot_feasible_domain_{name}_{analysis_case}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Abstract class : Analysis

@author: Angshuman Deb
"""

from abc import ABC, abstractmethod
import numpy as np
from .utility import Utility
import os


class Analysis(ABC):

    stampede_knl_n = 68
    stampede_skx_n = 48

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, analysis_type, structure, comp_env):
        self.analysis_type = analysis_type
        self.structure = structure
        self.comp_env = comp_env

    ####################################################################################################################
    # Abstract methods
    ####################################################################################################################

    @abstractmethod
    def setup(self, *args, **kwargs):
        pass

    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def run(self, *args, **kwargs):
        pass

    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def wrap_up(self, *args, **kwargs):
        pass

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def get_setup_dir_path(self):
        setup_dir_path = Utility.get_path(self.structure.model_work_dir_path,
                                          'Work_Dir', self.analysis_type + '_Setup')
        return setup_dir_path

    # ------------------------------------------------------------------------------------------------------------------

    def get_results_dir_path(self):
        results_dir_path = Utility.get_path(self.structure.model_work_dir_path,
                                            'Work_Dir', self.analysis_type + '_Results')
        return results_dir_path

    # ------------------------------------------------------------------------------------------------------------------

    def get_analysis_setup_info(self, analysis_case):
        analysis_setup_info = Utility.pickle_load_dict(
            Utility.get_path(
                self.get_setup_dir_path(), analysis_case, 'analysis_setup_info.pickle'
            )
        )
        return analysis_setup_info

    # ------------------------------------------------------------------------------------------------------------------

    def set_analysis_setup_info(self, analysis_case, analysis_setup_info):
        Utility.pickle_dump_dict(
            Utility.get_path(
                self.get_setup_dir_path(), analysis_case, 'analysis_setup_info.pickle'
            ), analysis_setup_info
        )
        return

    # ------------------------------------------------------------------------------------------------------------------

    def generate_batch_file(self, analysis_case, n_batch, n_job, pool_size_factor):

        structure = self.structure
        name = structure.name
        model_work_dir_path = structure.model_work_dir_path
        analysis_type = self.analysis_type
        comp_env = self.comp_env
        slurm_file_name = f'slurm_file_{name}_{analysis_type.lower()}_case_{analysis_case}'
        batch_file_name_prefix = f'batch_file_{name}_{analysis_type.lower()}_case_{analysis_case}'
        err_file_name_prefix = f'err_file_{name}_{analysis_type.lower()}_case_{analysis_case}'
        out_file_name_prefix = f'out_file_{name}_{analysis_type.lower()}_case_{analysis_case}'

        pool_size = int(getattr(self, comp_env + '_n') * pool_size_factor)
        pool_size = max(1, pool_size)

        for iter_batch in range(1, n_batch + 1):
            with open(f'{model_work_dir_path}/{batch_file_name_prefix}_batch_{iter_batch}', 'w') as batch_fid:
                batch_fid.write(f'#!/bin/bash\n')

                write_line = f'if [ ! -d "{model_work_dir_path}/Work_Dir/EO_Files" ]; '
                write_line += f'then mkdir "{model_work_dir_path}/Work_Dir/EO_Files"; fi;\n'
                batch_fid.write(write_line)

                batch_fid.write(f'for i in {{{n_job * (iter_batch - 1) + 1}..{n_job * iter_batch}}}; do\n')

                write_line = f'sbatch -J run$i '
                write_line += f'-e "{model_work_dir_path}/Work_Dir/EO_Files/{err_file_name_prefix}_job_$i" '
                write_line += f'-o "{model_work_dir_path}/Work_Dir/EO_Files/{out_file_name_prefix}_job_$i" '
                write_line += f'{slurm_file_name} $i {iter_batch} {pool_size}\n'
                batch_fid.write(write_line)

                batch_fid.write(f'done\n')
            os.system(f'chmod +x "{model_work_dir_path}/{batch_file_name_prefix}_batch_{iter_batch}"')
        return slurm_file_name

    # ------------------------------------------------------------------------------------------------------------------

    def generate_slurm_file(self, analysis_case, slurm_file_name, run_time_limit, allocation_name, n_batch, n_job):

        structure = self.structure
        model_work_dir_path = structure.model_work_dir_path
        analysis_type = self.analysis_type
        comp_env = self.comp_env

        with open(f'{model_work_dir_path}/{slurm_file_name}', 'w') as slurm_file_fid:
            slurm_file_fid.write(f'#!/bin/bash\n')
            slurm_file_fid.write(f'#SBATCH -N 1\n')
            n = getattr(self, comp_env + '_n')
            slurm_file_fid.write(f'#SBATCH -n {n}\n')
            if comp_env == 'stampede_knl':
                slurm_file_fid.write(f'#SBATCH -p normal\n')
            elif comp_env == 'stampede_skx':
                slurm_file_fid.write(f'#SBATCH -p skx-normal\n')
            slurm_file_fid.write(f'#SBATCH -t {run_time_limit}\n')
            slurm_file_fid.write(f'#SBATCH -A {allocation_name}\n')
            slurm_file_fid.write(f'\n')
            slurm_file_fid.write(f'module load python3\n')

            python_base_path = Utility.get_path(os.getcwd())
            write_line = f'python3 -c "import sys; sys.path.insert(1, \'{python_base_path}\'); '
            write_line += f'from run_on_stampede import run_on_stampede_{analysis_type.lower()} as run_on_stampede; '
            write_line += f'run_on_stampede(\'{analysis_case}\', {n_batch}, {n_job}, $1, $2, $3)"\n'
            slurm_file_fid.write(write_line)
        return

    # ------------------------------------------------------------------------------------------------------------------

    def get_analysis_info(self, analysis_case, list_file_name, n_batch, n_job, node_index, batch_index):

        list_file_path = Utility.get_path(self.get_setup_dir_path(), analysis_case, list_file_name)

        if node_index > n_job:
            if node_index % n_job == 0:
                node_index = n_job
            else:
                node_index = node_index % n_job

        analysis_list = Utility.read_numpy_array_from_txt_file(list_file_path, skiprows='find')

        if analysis_list.size != 0:
            tot_num_analysis = analysis_list.shape[0]
        else:
            tot_num_analysis = 0

        n_sim = tot_num_analysis

        analysis_per_batch = np.zeros(n_batch)
        analysis_per_batch[0: analysis_per_batch.size - 1] = np.floor(n_sim / n_batch)
        analysis_per_batch[analysis_per_batch.size - 1] = n_sim - (n_batch - 1) * np.floor(n_sim / n_batch)

        analysis_this_batch = analysis_per_batch[batch_index - 1]

        analysis_per_job = np.zeros(n_job)
        analysis_per_job[0: analysis_per_job.size - 1] = np.floor(analysis_this_batch / n_job)
        analysis_per_job[analysis_per_job.size - 1] = analysis_this_batch - (n_job - 1) * np.floor(
            analysis_this_batch / n_job)

        analysis_range_i = sum(analysis_per_batch[0: batch_index - 1]) + sum(analysis_per_job[0: node_index - 1]) + 1
        analysis_range_f = sum(analysis_per_batch[0: batch_index - 1]) + sum(analysis_per_job[0: node_index])

        analysis_range = np.arange(analysis_range_i, analysis_range_f + 1)

        return analysis_list.astype(int), analysis_range.astype(int), len(analysis_range)

    # ------------------------------------------------------------------------------------------------------------------

    def setup_non_local_run(self, analysis_case, n_batch, n_job,
                            pool_size_factor, run_time_limit, allocation_name):
        # Write slurm job file and batch files:
        slurm_file_name = self.generate_batch_file(analysis_case, n_batch, n_job, pool_size_factor)
        self.generate_slurm_file(analysis_case, slurm_file_name, run_time_limit, allocation_name, n_batch, n_job)
        return

    # ------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 20 12:36:55 2019

@author: Angshuman Deb
"""

from abc import ABC, abstractmethod
from .structure import Structure
from .utility import Utility
import numpy as np
import scipy.io as sio
import os
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.art3d as art3d
import random
from benedict import benedict
from functools import partial
from collections.abc import Sequence
from pathos.multiprocessing import ProcessPool as Pool


class EDP(ABC):
    
    ####################################################################################################################
    # Constructor
    ####################################################################################################################
    
    def __init__(self, structure, tag, recorder_file_storage, haz_req, small_value, large_value):
        self.structure = structure
        self.tag = tag
        self.recorder_file_storage = recorder_file_storage
        self.haz_req = haz_req
        self.small_value = small_value
        self.large_value = large_value

    ####################################################################################################################
    # Setters (instance methods)
    ####################################################################################################################
    
    def set_haz_req(self, haz_req):
        self.haz_req = haz_req

    ####################################################################################################################
    # Static methods
    ####################################################################################################################

    @staticmethod
    def collect_parallel(tag, results_dir_path, edp_strs, run_case):

        for_which = list(run_case.astype(str))
        run_case_str = ' '.join(for_which)

        target_work_dir_path = Utility.get_path(results_dir_path, *for_which, f'EDP_{tag}_Results')
        nltha_status_file_path = Utility.get_path(results_dir_path, *for_which, 'NLTHA_STATUS.txt')
        status = 'FAIL'
        if os.path.isfile(nltha_status_file_path):
            with open(nltha_status_file_path, 'r') as fid:
                status = fid.readline().strip()

        to_return = dict()
        for edp_str in edp_strs:
            edp_file_name = Utility.get_path(target_work_dir_path, f'edp_{tag}_{edp_str}.txt')
            if os.path.isfile(edp_file_name) and status == 'SUCCESS':
                with open(Utility.get_path(
                        target_work_dir_path, f'edp_{tag}_{edp_str}.txt'), 'r') as fid:
                    edp_val = float(fid.readline().strip('\n'))
            else:
                edp_val = np.inf

            to_return[f'{edp_str}'] = edp_val

        print(f'Collecting edp {tag} for case {run_case_str} complete!')

        return f'{run_case_str}', to_return

    ####################################################################################################################
    # Abstract methods
    ####################################################################################################################

    @abstractmethod
    def write_evaluate_edp_and_helpers(self, *args, **kwargs):
        pass

    @abstractmethod
    def generate_recorder(self, *args, **kwargs):
        pass

    @abstractmethod
    def get_edp_strings(self, *args, **kwargs):
        pass

    ####################################################################################################################
    # Public functionalities
    ####################################################################################################################

    def collect(self, analysis_case, pool_size, **kwargs):
        analysis_list = kwargs.get('analysis_list', 'analysis_list_original.txt')
        structure = self.structure
        setup_dir_path = Utility.get_path(structure.model_work_dir_path, 'Work_Dir', 'NLTHA_Setup')

        if isinstance(analysis_list, str):
            list_file_path = Utility.get_path(setup_dir_path, analysis_case, analysis_list)
            analysis_list = Utility.read_numpy_array_from_txt_file(list_file_path, skiprows='find').astype(int)

        if analysis_list.size == 0:
            return

        results_dir_path = Utility.get_path(structure.model_work_dir_path, 'Work_Dir', 'NLTHA_Results')
        analysis_range = np.arange(1, analysis_list.shape[0] + 1)
        to_pass = [[analysis_list[analysis_index - 1, :] for analysis_index in analysis_range], ]

        design_num_list = list(np.unique(analysis_list[:, 1]).astype(str))

        edp_strs = self.get_edp_strings(list(analysis_list[0, :].astype(str)))
        tag = self.tag
        to_run = partial(EDP.collect_parallel, tag, results_dir_path, edp_strs)

        if pool_size > 1:
            my_pool = Pool(pool_size)
            edp_results_this = dict(my_pool.map(to_run, *to_pass))
            my_pool.close()
            my_pool.join()
            my_pool.clear()
        else:
            edp_results_this = dict(list(map(to_run, *to_pass)))

        for key in edp_results_this.keys():
            for edp_str in edp_strs:
                if edp_results_this[key][f'{edp_str}'] == np.inf:
                    warning_str = f"Warning: Invalid edp_{tag}_{edp_str} " \
                                  f"found for for_which = {key}!"
                    print(warning_str)
                    for_which = key.split()
                    nltha_status_file_path = Utility.get_path(
                        results_dir_path, *for_which, 'NLTHA_STATUS.txt')
                    with open(nltha_status_file_path, 'w') as fid:
                        fid.write('FAIL\n')
                if edp_results_this[key][f'{edp_str}'] == 0:
                    print(f"Warning: edp_{tag}_{edp_str} = 0.0 found for for_which = {key}!")
                    edp_results_this[key][f'{edp_str}'] = self.small_value

        for design_num in design_num_list:
            target_work_dir_path = Utility.get_path(results_dir_path.replace('_Results', '_EDP'), analysis_case,
                                                    design_num)

            if not os.path.isdir(target_work_dir_path):
                os.makedirs(target_work_dir_path)

            edp_results_file_path = Utility.get_path(target_work_dir_path, 'edp_results.pickle')

            if os.path.isfile(edp_results_file_path):
                edp_results = benedict(Utility.pickle_load_dict(edp_results_file_path))
            else:
                edp_results = benedict()

            edp_results_this_design_num = {key: edp_results_this[key] for key in edp_results_this.keys() if
                                           key.split()[Structure.get_dir_level_index('Design_Num')] == design_num}

            edp_results.merge({tag: edp_results_this_design_num})
            Utility.pickle_dump_dict(edp_results_file_path, edp_results)
        return

    def get_rec_save_dir_path(self):
        if self.recorder_file_storage == 'shared':
            rec_save_dir_path = Utility.get_path(self.structure.structural_analysis_platform.nltha_rec_dir_name, 'EDP')
        else:
            rec_save_dir_path = Utility.get_path(self.structure.structural_analysis_platform.nltha_rec_dir_name, f'EDP_{self.tag}')
        return rec_save_dir_path

    def compute_demand_hazard_integral(self, for_which, haz_lev_list, im, n_gm_list=None, **kwargs):
        delta_input = kwargs.get('delta_input', np.array([]))
        delta_input = np.array(delta_input).flatten()  # 1-d array
        min_max_scale_fac = kwargs.get('min_max_scale_fac', [1, 1])  # list

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        mrp_list = list()
        n_gm_list_send = list()

        for haz_lev in haz_lev_list:
            _, _, mrp, n_gm = im.get_gms_results(for_which + [haz_lev])
            mrp_list.append(mrp)
            n_gm_list_send.append(n_gm)

        if isinstance(n_gm_list, Sequence):
            n_gm_list_send = n_gm_list

        seismic_hazard_curve, _ = im.get_psha_results(for_which)

        inv_seismic_hazard_at_haz_lev_list = list()
        for itr in range(len(haz_lev_list)):
            inv_seismic_hazard_at_haz_lev_list.append(
                seismic_hazard_curve[np.argmin(np.abs(seismic_hazard_curve[:, 1] - (1 / mrp_list[itr]))), 0])

        im_vals = seismic_hazard_curve[:, 0]
        im_vals_integration = np.exp(np.mean(np.log(np.column_stack([im_vals[:-1], im_vals[1:]])), axis=1))
        nu_im = seismic_hazard_curve[:, 1]
        dnu_im = np.abs(np.diff(nu_im))

        to_return = benedict()
        edp_strs = self.get_edp_strings(for_which)
        for edp_str in edp_strs:
            dhc_save, nu_edp_im_save = self._compute_demand_hazard_integral(for_which, haz_lev_list,
                                                                            n_gm_list_send,
                                                                            inv_seismic_hazard_at_haz_lev_list,
                                                                            im_vals_integration, dnu_im,
                                                                            edp_str, delta_input,
                                                                            min_max_scale_fac)

            to_return.merge(
                {f'{edp_str}': {'demand_hazard_curve': dhc_save,
                                'demand_hazard_curve_deagg_im': nu_edp_im_save
                                }
                 }
            )

        to_return.merge(
            {'seismic_hazard_curve': np.column_stack(
                [im_vals_integration,
                 np.exp(np.mean(np.log(np.column_stack([nu_im[:-1], nu_im[1:]])), axis=1)),
                 dnu_im]
            )
            }
        )

        to_return.merge({'mrp_list': mrp_list})

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def get_inv_demand_hazard(self, mrp, for_which, edp_str):
        dhc, _, _, _ = self.get_psdemha_results(for_which, edp_str)
        return dhc[np.argmin(np.abs(dhc[:, 1] - 1 / mrp)), 0]

    # ------------------------------------------------------------------------------------------------------------------

    def get_demand_hazard_deagg_im(self, mrp, for_which, edp_str):
        dhc, dhc_deagg_im, _, _ = self.get_psdemha_results(for_which, edp_str)
        ind = np.argmin(np.abs(dhc[:, 1] - 1 / mrp))
        # return 1-d array
        return dhc_deagg_im[ind, :] / dhc[ind, 1]

    # ------------------------------------------------------------------------------------------------------------------

    def get_psdemha_results(self, for_which, edp_str=None):
        model_work_dir_path = self.structure.model_work_dir_path
        dir_category = 'PSDemHA_Results'
        file_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category,
                                     *Structure.get_modified_for_which(for_which, 'Design_Num'), 'psdemha_results.pickle')
        psdemha_results = Utility.pickle_load_dict(file_path)
        if len(for_which) == (Structure.get_dir_level_index('Design_Num') + 1) or edp_str is None:
            return psdemha_results[self.tag]

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        run_case_str = ' '.join(for_which)
        psdemha_results_for_which = psdemha_results[self.tag][run_case_str]

        to_return = list()
        to_return.append(psdemha_results_for_which[f'{edp_str}']['demand_hazard_curve'])
        to_return.append(psdemha_results_for_which[f'{edp_str}']['demand_hazard_curve_deagg_im'])
        to_return.append(psdemha_results_for_which['seismic_hazard_curve'])
        to_return.append(psdemha_results_for_which['mrp_list'])

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def get_edp_at_hazard_level(self, for_which, haz_lev_list, n_gm_list, edp_str):
        tag = self.tag
        edp_vals = list()

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        edp_results = self.get_edp_results(for_which)

        random.seed(999)
        for haz_lev in haz_lev_list:
            temp = [edp_results[key][f'{edp_str}'] for key in edp_results.keys() if
                    key.split()[Structure.get_dir_level_index('Hazard_Level_Num')] == haz_lev and
                    Structure.get_modified_for_which(key.split(), dir_level_to_seek) == for_which]
            temp = [temp_content for temp_content in temp if temp_content > 0.0]
            temp = [temp_content for temp_content in temp if temp_content < np.inf]
            if len(temp) != n_gm_list[haz_lev_list.index(haz_lev)]:
                warning_str = f"Warning: Number of edp_{tag}_{edp_str} values found for haz_lev = {haz_lev} " \
                              f"doesn't match " \
                              f"n_gm = {n_gm_list[haz_lev_list.index(haz_lev)]} for for_which = {for_which}!"
                print(warning_str)
                try:
                    temp = random.sample(temp, n_gm_list[haz_lev_list.index(haz_lev)])
                except ValueError:
                    pass
            edp_vals.append(np.array(temp))

        return edp_vals

    # ------------------------------------------------------------------------------------------------------------------

    def get_conditional_demand_params(self, for_which, haz_lev_list, n_gm_list,
                                      inv_seismic_hazard_at_haz_lev_list, im_vals_integration,
                                      edp_str):
        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        edp_vals = self.get_edp_at_hazard_level(for_which, haz_lev_list, n_gm_list, edp_str)

        haz_req = self.haz_req
        fit_dist = haz_req['fit_dist']
        transf_on_dist_params_list = haz_req['transf_on_dist_params_list']
        transf_to_dist_params_list = haz_req['transf_to_dist_params_list']
        interp_exterp_function_list = haz_req['interp_exterp_function_list']

        dist_params = np.zeros((len(haz_lev_list), len(transf_on_dist_params_list)))
        transf_params = np.zeros((len(haz_lev_list), len(transf_on_dist_params_list)))

        for itr in range(len(haz_lev_list)):
            dist_params[itr, :] = Utility.fit_dist_to_data(fit_dist, edp_vals[itr])

        for itr in range(transf_params.shape[1]):
            transf_params[:, itr] = transf_on_dist_params_list[itr](*dist_params.T)

        transf_params_interp_exterp = np.zeros((len(im_vals_integration), transf_params.shape[1]))
        dist_params_interp_exterp = np.zeros((len(im_vals_integration), transf_params.shape[1]))

        for itr in range(transf_params.shape[1]):
            transf_params_interp_exterp[:, itr] = interp_exterp_function_list[itr](
                inv_seismic_hazard_at_haz_lev_list, transf_params[:, itr], im_vals_integration)

        for itr in range(transf_params.shape[1]):
            dist_params_interp_exterp[:, itr] = transf_to_dist_params_list[itr](*transf_params_interp_exterp.T)

        to_return = dict()
        to_return['dist_params'] = dist_params
        to_return['transf_params'] = transf_params
        to_return['dist_params_interp_exterp'] = dist_params_interp_exterp
        to_return['transf_params_interp_exterp'] = transf_params_interp_exterp
        to_return['edp_vals'] = edp_vals

        return to_return

    # ------------------------------------------------------------------------------------------------------------------

    def _compute_demand_hazard_integral(self, for_which, haz_lev_list, n_gm_list, inv_seismic_hazard_at_haz_lev_list,
                                        im_vals_integration, dnu_im, edp_str, delta_input, min_max_scale_fac):

        fit_dist = self.haz_req['fit_dist']
        conditional_demand_params = self.get_conditional_demand_params(for_which,
                                                                       haz_lev_list,
                                                                       n_gm_list,
                                                                       inv_seismic_hazard_at_haz_lev_list,
                                                                       im_vals_integration, edp_str)

        dist_params_interp_exterp = conditional_demand_params['dist_params_interp_exterp']
        edp_vals = conditional_demand_params['edp_vals']

        if delta_input.size == 0:
            min_edp = min_max_scale_fac[0] * min([min(temp) for temp in edp_vals])
            max_edp = min_max_scale_fac[1] * max([max(temp) for temp in edp_vals])
            delta_input = np.logspace(np.log10(min_edp), np.log10(max_edp), 1000, endpoint=True)
        delta_input = delta_input.reshape(len(delta_input), 1)  # 2-d vector

        nu_edp_im = (1 - fit_dist.cdf(delta_input, *dist_params_interp_exterp.T)) * dnu_im
        nu_edp = np.sum(nu_edp_im, axis=1)
        dhc = np.column_stack([delta_input, nu_edp])

        dhc_save = dhc[dhc[:, 1] != 0, :]
        nu_edp_im_save = nu_edp_im[dhc[:, 1] != 0, :]

        return dhc_save, nu_edp_im_save

    # ------------------------------------------------------------------------------------------------------------------

    def get_edp_results(self, for_which):
        model_work_dir_path = self.structure.model_work_dir_path
        tag = self.tag
        dir_category = 'NLTHA_EDP'
        dir_level_to_seek = 'Design_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        edp_results = Utility.pickle_load_dict(
            Utility.get_path(
                model_work_dir_path, 'Work_Dir', dir_category, *for_which, 'edp_results.pickle'
            )
        )[tag]

        return edp_results

    # ------------------------------------------------------------------------------------------------------------------

    def find_anomalous_run_cases(self, for_which, threshold):

        dir_level_to_seek = 'Design_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        edp_results = self.get_edp_results(for_which)
        run_cases = [np.array(key.split()).astype(int) for key in edp_results.keys()
                     if np.any(np.array(list(edp_results[key].values())) > threshold)]
        run_cases = np.array(run_cases)

        if run_cases.shape[0] == 0:
            return run_cases.reshape(0, len(Structure.get_all_dir_levels()))
        else:
            return run_cases

    # ------------------------------------------------------------------------------------------------------------------

    def plot_conditional_demand_regression_model(self, for_which, edp_str,
                                                 haz_lev_list, im, **kwargs):

        figkwargs = kwargs.get('figkwargs', dict())
        lc = kwargs.get('lc', 'blue')
        mc = kwargs.get('mc', 'red')
        lw = kwargs.get('lw', 1)
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        im_lim = kwargs.get('im_lim', list())
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        fig_ax = kwargs.get('fig_ax', None)

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        mrp_list = list()
        n_gm_list = list()

        for haz_lev in haz_lev_list:
            _, _, mrp, n_gm = im.get_gms_results(for_which + [haz_lev])
            mrp_list.append(mrp)
            n_gm_list.append(n_gm)

        haz_req = self.haz_req
        num_params = len(haz_req['interp_exterp_function_list'])

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig, axs = plt.subplots(1, num_params, **figkwargs)  # also accepts tight_layout=bool
        else:
            fig = fig_ax[0]
            axs = fig_ax[1]
        for ax in axs:
            ax.minorticks_on()
            ax.grid(True, which="major", alpha=grid_alpha)
            ax.grid(True, which="minor", alpha=minor_grid_alpha)

        inv_seismic_hazard_at_haz_lev_list = list()
        for itr in range(len(haz_lev_list)):
            inv_seismic_hazard_at_haz_lev_list.append(im.get_inv_seismic_hazard(mrp_list[itr], for_which))

        seismic_hazard_curve, _ = im.get_psha_results(for_which)
        im_vals = seismic_hazard_curve[:, 0]
        im_vals_integration = np.exp(np.mean(np.log(np.column_stack([im_vals[:-1], im_vals[1:]])), axis=1))

        if len(im_lim) == 0:
            im_lim = [im_vals_integration[0], im_vals_integration[-1]]

        conditional_demand_params = self.get_conditional_demand_params(for_which,
                                                                       haz_lev_list,
                                                                       n_gm_list,
                                                                       inv_seismic_hazard_at_haz_lev_list,
                                                                       im_vals_integration, edp_str)

        transf_params = conditional_demand_params['transf_params']
        transf_params_interp_exterp = conditional_demand_params['transf_params_interp_exterp']

        for itr in range(transf_params.shape[1]):
            axs[itr].plot(inv_seismic_hazard_at_haz_lev_list, transf_params[:, itr], 'o', color=mc)
            axs[itr].plot(*Utility.get_xy_within_lims(im_vals_integration,
                                                      transf_params_interp_exterp[:, itr],
                                                      im_lim, [-np.inf, np.inf]), '-', color=lc, linewidth=lw)
            axs[itr].set_xlim(im_lim)

        export_mat_dict = dict()
        export_mat_dict[f'inv_seismic_hazard_at_haz_lev_list'] = inv_seismic_hazard_at_haz_lev_list
        export_mat_dict[f'im_vals_integration'] = im_vals_integration
        export_mat_dict[f'transf_params'] = transf_params
        export_mat_dict[f'transf_params_interp_exterp'] = transf_params_interp_exterp

        if save_mat:
            structure = self.structure
            name = structure.name
            tag = self.tag
            for_which_str = '_'.join(for_which)
            file_name = f'plot_conditional_demand_regression_model_{name}_{for_which_str}_' \
                        f'edp_{tag}_{edp_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, axs, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_dhc(self, for_which, edp_str, **kwargs):

        figkwargs = kwargs.get('figkwargs', dict())
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        grid_alpha = kwargs.get('grid_alpha', 0.5)
        minor_grid_alpha = kwargs.get('minor_grid_alpha', 0.0)
        lc = kwargs.get('lc', 'black')
        lw = kwargs.get('lw', 1.2)
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.gca()
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        ax.minorticks_on()
        ax.grid(True, which="major", alpha=grid_alpha)
        ax.grid(True, which="minor", alpha=minor_grid_alpha)

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        dhc = self.get_psdemha_results(for_which, edp_str)[0]
        ax.loglog(dhc[:, 0], dhc[:, 1], color=lc, linewidth=lw)

        export_mat_dict = dict()
        export_mat_dict['dhc'] = dhc

        if save_mat:
            structure = self.structure
            name = structure.name
            tag = self.tag
            for_which_str = '_'.join(for_which)
            file_name = f'plot_dhc_{name}_{for_which_str}_edp_{tag}_{edp_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------

    def plot_conditional_demand_model_3d(self, for_which, edp_str, haz_lev_list, im, **kwargs):

        figkwargs = kwargs.get('figkwargs', dict())
        mc = kwargs.get('mc', 'red')
        ms = kwargs.get('ms', 2)
        lw = kwargs.get('lw', 1)
        lc = kwargs.get('lc', 'blue')
        save_dir_path = kwargs.get('save_dir_path', Utility.get_path(os.getcwd()))
        save_mat = kwargs.get('save_mat', False)
        edp_lim = kwargs.get('edp_lim', [self.small_value, self.large_value])
        im_lim = kwargs.get('im_lim', list())
        n_bins = kwargs.get('n_bins', 10)
        first_pctl = kwargs.get('first_pctl', 0.025)
        second_pctl = kwargs.get('second_pctl', 0.50)
        third_pctl = kwargs.get('third_pctl', 0.975)
        minor_grid = kwargs.get('minor_grid', False)
        patch_color = kwargs.get('patch_color', 'cyan')
        patch_alpha = kwargs.get('patch_alpha', 0.25)
        n_cont = kwargs.get('n_cont', 100)
        fig_ax = kwargs.get('fig_ax', None)

        if fig_ax is None or np.array(fig_ax, dtype='object').any() is None:
            fig = plt.figure(**figkwargs)
            ax = fig.add_subplot(projection='3d')
        else:
            fig = fig_ax[0]
            ax = fig_ax[1]
        if minor_grid:
            ax.minorticks_on()
        else:
            ax.minorticks_off()

        dir_level_to_seek = 'Model_Realization_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)

        mrp_list = list()
        n_gm_list = list()

        for haz_lev in haz_lev_list:
            _, _, mrp, n_gm = im.get_gms_results(for_which + [haz_lev])
            mrp_list.append(mrp)
            n_gm_list.append(n_gm)

        haz_req = self.haz_req
        fit_dist = haz_req['fit_dist']

        inv_seismic_hazard_at_haz_lev_list = list()
        for itr in range(len(haz_lev_list)):
            inv_seismic_hazard_at_haz_lev_list.append(im.get_inv_seismic_hazard(mrp_list[itr], for_which))

        seismic_hazard_curve, _ = im.get_psha_results(for_which)
        im_vals = seismic_hazard_curve[:, 0]
        im_vals_integration = np.exp(np.mean(np.log(np.column_stack([im_vals[:-1], im_vals[1:]])), axis=1))

        if len(im_lim) == 0:
            im_lim = [im_vals_integration[0], im_vals_integration[-1]]

        conditional_demand_params = self.get_conditional_demand_params(for_which,
                                                                       haz_lev_list,
                                                                       n_gm_list,
                                                                       inv_seismic_hazard_at_haz_lev_list,
                                                                       im_vals_integration, edp_str)

        edp_vals = conditional_demand_params['edp_vals']
        dist_params_interp_exterp = conditional_demand_params['dist_params_interp_exterp']
        dist_params = conditional_demand_params['dist_params']

        first_pctl_values = fit_dist.ppf(first_pctl, *dist_params_interp_exterp.T)
        second_pctl_values = fit_dist.ppf(second_pctl, *dist_params_interp_exterp.T)
        second_pctl_values_haz_lev = fit_dist.ppf(second_pctl, *dist_params.T)
        third_pctl_values = fit_dist.ppf(third_pctl, *dist_params_interp_exterp.T)

        ax.plot(*Utility.get_xy_within_lims(first_pctl_values, im_vals_integration, edp_lim, im_lim), 0,
                '-.', color=lc, linewidth=lw)
        ax.plot(*Utility.get_xy_within_lims(second_pctl_values, im_vals_integration, edp_lim, im_lim), 0,
                '-.', color=lc, linewidth=lw)
        ax.plot(second_pctl_values_haz_lev, inv_seismic_hazard_at_haz_lev_list, 0, 'd', color=mc,
                markeredgecolor='black', markersize=ms * 2)
        ax.plot(*Utility.get_xy_within_lims(third_pctl_values, im_vals_integration, edp_lim, im_lim), 0,
                '-.', color=lc, linewidth=lw)

        export_mat_dict = dict()

        for itr in range(len(haz_lev_list)):
            x = edp_vals[itr]
            export_mat_dict[f'edp_vals_haz_lev_{haz_lev_list[itr]}'] = x
            y = np.ones(x.size) * inv_seismic_hazard_at_haz_lev_list[itr]
            ax.plot(*Utility.get_xy_within_lims(x, y, edp_lim, im_lim), 0, '.', color=mc, markersize=ms)
            hist, bin_edges = np.histogram(x, bins=n_bins, density=True)
            export_mat_dict[f'hist_haz_lev_{haz_lev_list[itr]}'] = hist
            export_mat_dict[f'bin_edges_haz_lev_{haz_lev_list[itr]}'] = bin_edges
            for i_bin in range(len(hist)):
                rect = plt.Rectangle((bin_edges[i_bin], 0),
                                     bin_edges[i_bin + 1] - bin_edges[i_bin],
                                     hist[i_bin],
                                     edgecolor='black', facecolor='None', linewidth=lw / 2)
                ax.add_patch(rect)
                art3d.pathpatch_2d_to_3d(rect, z=inv_seismic_hazard_at_haz_lev_list[itr], zdir="y")
            x_cont = np.logspace(np.log10(edp_lim[0]), np.log10(edp_lim[-1]), n_cont)
            y_cont = np.ones(x_cont.size) * inv_seismic_hazard_at_haz_lev_list[itr]
            z_cont = fit_dist.pdf(x_cont, *dist_params[itr, :])
            ax.plot(x_cont, y_cont, z_cont, '-', color=lc, linewidth=lw)
            export_mat_dict[f'cont_pdf_curves_haz_lev_{haz_lev_list[itr]}'] = np.column_stack([x_cont, z_cont])

        rect = plt.Rectangle((edp_lim[0], inv_seismic_hazard_at_haz_lev_list[0]),
                             edp_lim[-1] - edp_lim[0],
                             inv_seismic_hazard_at_haz_lev_list[-1] - inv_seismic_hazard_at_haz_lev_list[0],
                             color=patch_color, alpha=patch_alpha)
        ax.add_patch(rect)
        art3d.pathpatch_2d_to_3d(rect, z=0)

        zlim = ax.get_zlim()
        ax.set_zlim([0, zlim[-1]])
        ax.set_xlim(edp_lim)
        ax.set_ylim(im_lim)

        export_mat_dict[f'inv_seismic_hazard_at_haz_lev_list'] = inv_seismic_hazard_at_haz_lev_list
        export_mat_dict[f'im_vals_integration'] = im_vals_integration
        export_mat_dict[f'first_pctl_values'] = first_pctl_values
        export_mat_dict[f'second_pctl_values'] = second_pctl_values
        export_mat_dict[f'second_pctl_values_haz_lev'] = second_pctl_values_haz_lev
        export_mat_dict[f'third_pctl_values'] = third_pctl_values
        export_mat_dict[f'fit_dist_name'] = fit_dist.name
        export_mat_dict[f'seismic_hazard_curve'] = seismic_hazard_curve
        export_mat_dict[f'dist_params_interp_exterp'] = dist_params_interp_exterp
        export_mat_dict[f'dist_params'] = dist_params

        if save_mat:
            structure = self.structure
            name = structure.name
            tag = self.tag
            for_which_str = '_'.join(for_which)
            file_name = f'plot_conditional_demand_model_3d_{name}_{for_which_str}_' \
                        f'edp_{tag}_{edp_str}.mat'
            sio.savemat(Utility.get_path(save_dir_path, file_name), export_mat_dict)

        return fig, ax, export_mat_dict

    # ------------------------------------------------------------------------------------------------------------------


class FrameMaxDeformation(EDP, ABC):
    def __init__(self, max_what, frame_structure, tag, recorder_file_storage, haz_req, small_value, large_value):
        super().__init__(frame_structure, tag, recorder_file_storage, haz_req, small_value, large_value)
        self.max_what = max_what
        self.resp = 'deformation'
        self.normalize_with = 1.0
        self.resp_index = 1

    @abstractmethod
    def write_generate_edp_recorders(self, for_which, rec_gen_file_open_mode):
        pass

    def write_evaluate_edp_and_helpers(self):
        max_what = self.max_what
        tag = self.tag
        resp = self.resp
        resp_index = self.resp_index
        self.structure.structural_analysis_platform.write_evaluate_edp_and_helpers(
            tag, max_what, resp, resp_index)
        return

    def generate_recorder(self, for_which, **kwargs):
        rec_gen_file_open_mode = kwargs.get('rec_gen_file_open_mode', 'a+')
        gen_rec = kwargs.get('gen_rec', True)

        frame_structure = self.structure
        model_work_dir_path = frame_structure.model_work_dir_path
        dir_category = 'NLTHA_Results'
        dir_level_to_seek = 'Ground_Motion_Rec_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        if gen_rec:
            self.write_generate_edp_recorders(for_which, rec_gen_file_open_mode)
        frame_structure.structural_analysis_platform.write_evaluate_edps(work_dir_path, self, for_which, gen_rec, rec_gen_file_open_mode)
        return


class MaxColRebarStrain(FrameMaxDeformation):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, max_what, frame_structure, tag, recorder_file_storage, **kwargs):
        haz_req = kwargs.get('haz_req', {})
        small_value = kwargs.get('small_value', 1.0e-10)
        large_value = kwargs.get('large_value', 1.0e0)
        super().__init__(max_what, frame_structure, tag, recorder_file_storage, haz_req, small_value, large_value)
        self.resp = 'strain'
        self.resp_index = 2

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def write_generate_edp_recorders(self, for_which, rec_gen_file_open_mode):
        frame_structure = self.structure
        model_work_dir_path = frame_structure.model_work_dir_path
        dir_category = 'NLTHA_Results'
        rec_save_dir_path = self.get_rec_save_dir_path()
        dir_level_to_seek = 'Ground_Motion_Rec_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        frame_structure.write_col_rebar_strain_recorder(for_which, work_dir_path,
                                                        rec_gen_file_open_mode, rec_save_dir_path)
        return

    def get_edp_strings(self, for_which):
        model_params = self.structure.model_params
        num_cols_total = model_params['num_cols_total']
        col_edge_list = model_params['col_edge_list']
        edp_strs = []
        for i_col in range(1, num_cols_total + 1):
            for i_edge in col_edge_list:
                col_id = i_col
                edge_id = i_edge
                edp_strs += [f'col_{col_id}_edge_{edge_id}']
        return edp_strs

    # ------------------------------------------------------------------------------------------------------------------


class MaxSpringDeformation(FrameMaxDeformation):

    ####################################################################################################################
    # Constructor
    ####################################################################################################################

    def __init__(self, spring_type, max_what, frame_structure, tag, recorder_file_storage, **kwargs):
        haz_req = kwargs.get('haz_req', {})
        normalize_with = kwargs.get('normalize_with', 1.0)
        small_value = kwargs.get('small_value', 0.)
        large_value = kwargs.get('large_value', 1.0e99)
        resp_index = kwargs.get('resp_index', 1)
        super().__init__(max_what, frame_structure, tag, recorder_file_storage, haz_req, small_value, large_value)
        self.spring_type = spring_type
        self.normalize_with = normalize_with
        self.resp_index = resp_index

    ####################################################################################################################
    # Public functionalities (instance methods)
    ####################################################################################################################

    def write_generate_edp_recorders(self, for_which, rec_gen_file_open_mode):
        frame_structure = self.structure
        model_work_dir_path = frame_structure.model_work_dir_path
        dir_category = 'NLTHA_Results'
        rec_save_dir_path = self.get_rec_save_dir_path()
        dir_level_to_seek = 'Ground_Motion_Rec_Num'
        for_which = Structure.get_modified_for_which(for_which, dir_level_to_seek)
        work_dir_path = Utility.get_path(model_work_dir_path, 'Work_Dir', dir_category, Utility.get_path(*for_which))
        frame_structure.write_spring_deformation_recorder(for_which, self.spring_type, work_dir_path,
                                                          rec_gen_file_open_mode, rec_save_dir_path)
        return

    def get_edp_strings(self, for_which):
        frame_structure = self.structure
        spring_type = self.spring_type
        edp_strs = []
        num_springs = len(frame_structure.get_spring_elem_tags(for_which, spring_type))
        for i_spring in range(1, num_springs + 1):
            spring_id = i_spring
            edp_strs += [f'{spring_type}_{spring_id}']
        return edp_strs

    # ------------------------------------------------------------------------------------------------------------------
